# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS


class PAMR(OLPS):
    """
    This class implements the Passive Aggressive Mean Reversion strategy. It is reproduced with
    modification from the following paper:
    `Li, B., Zhao, P., Hoi, S.C., & Gopalkrishnan, V. (2012). PAMR: Passive aggressive mean
    reversion strategy for portfolio selection. Machine Learning, 87, 221-258.
    <https://link.springer.com/content/pdf/10.1007%2Fs10994-012-5281-z.pdf>`_

    Passive Aggressive Mean Reversion strategy switches between a passive and an aggressive mean
    reversion strategy based on epsilon, a measure of sensitivity to the market,
    and hyperparameter C, which denotes the aggressiveness of reverting to a particular strategy.
    """

    def __init__(self, optimization_method, epsilon=0.5, agg=10):
        """
        Initializes Passive Aggressive Mean Reversion with the given epsilon, aggressiveness,
        and optimzation method.

        :param optimization_method: (int) 0 for PAMR, 1 for PAMR-1, 2 for PAMR-2. All three methods
                                          tend to return similar values.
        :param epsilon: (float) Sensitivity to the market with range of [0, inf). Because the epsilon
                                is considered a threshold and daily returns typically are around 1.
                                It is suggested to keep the range of [0, 1.5]. Typically, the returns
                                are highest with either a value of 0 or 1 for epsilon. 0 indicates
                                active mean reversion for all periods, and 1 indicates passive mean
                                reversion for daily returns below 1.
        :param agg: (float) Aggressiveness to mean reversion with range [0, inf). Aggressiveness
                            does not have much of an impact as epsilon. Typically, 100 has the highest
                            returns for PAMR-1 and 10000 has the highest returns for PAMR-2.
        """
        self.epsilon = epsilon
        self.agg = agg
        self.optimization_method = optimization_method
        super().__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        super(PAMR, self)._initialize(asset_prices, weights, resample_by)

        # Check that epsilon is greater than 0.
        if self.epsilon < 0:
            raise ValueError("Epsilon values must be greater than 0")

        # Check that aggressiveness is greater than 0.
        if self.agg < 0:
            raise ValueError("Aggressiveness values must be greater than 0")

        # Check that optimization method is either 0, 1, or 2.
        if self.optimization_method not in [0, 1, 2]:
            raise ValueError("Optimization method must be either 0, 1, or 2.")

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # First prediction returns the same weights.
        if time == 0:
            return self.weights

        # Prepare for calculations.
        current_relative_return = self.relative_return[time]

        # Calculate loss function.
        loss = max(0, np.dot(self.weights, current_relative_return) - self.epsilon)

        # Calculate the adjusted market change.
        adjusted_market_change = current_relative_return - self._uniform_weight() \
                                 * np.mean(current_relative_return)

        # Calculate the norm of the adjusted market change.
        diff_norm = np.linalg.norm(adjusted_market_change)

        # PAMR method looks to passively perform mean reversion.
        if self.optimization_method == 0:
            tau = loss / (diff_norm ** 2)

        # PAMR-1 introduces a slack variable for tradeoff between epsilon and C.
        elif self.optimization_method == 1:
            tau = min(self.agg, loss / (diff_norm ** 2))

        # PAMR-2 introduces a quadratic slack variable for tradeoff between epsilon and C.
        else:
            tau = loss / (diff_norm ** 2 + 1 / (2 * self.agg))

        # Calculate new weights.
        new_weights = self.weights - tau * adjusted_market_change

        # Project new weights to simplex domain.
        new_weights = self._simplex_projection(new_weights)
        return new_weights



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import pairwise_distances
from scipy.cluster.hierarchy import fcluster
from scipy.cluster.hierarchy import linkage as scipy_linkage, dendrogram
from scipy.spatial.distance import squareform
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators
from famlafl.portfolio_optimization.risk_metrics import RiskMetrics
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators


class HierarchicalEqualRiskContribution:
    """
    This class implements the Hierarchical Equal Risk Contribution (HERC) algorithm and it's extended components mentioned in the
    following papers: `Raffinot, Thomas, The Hierarchical Equal Risk Contribution Portfolio (August 23,
    2018). <https://ssrn.com/abstract=3237540>`_; and `Raffinot, Thomas, Hierarchical Clustering Based Asset Allocation (May 2017)
    <https://ssrn.com/abstract=2840729>`_;

    While the vanilla Hierarchical Risk Parity algorithm uses only the variance as a risk measure for assigning weights, the HERC
    algorithm proposed by Raffinot, allows investors to use other risk metrics like Standard Deviation, Expected Shortfall and
    Conditional Drawdown at Risk.
    """

    UniqueColors = ['darkred', 'deepskyblue', 'springgreen', 'darkorange', 'deeppink', 'slateblue', 'navy', 'blueviolet',
                    'pink', 'darkslategray']
    UnclusteredColor = "#808080"
    def __init__(self, confidence_level=0.05):
        """
        Initialise.

        :param confidence_level: (float) The confidence level (alpha) used for calculating expected shortfall and conditional
                                         drawdown at risk.
        """

        self.weights = list()
        self.clusters = None
        self.ordered_indices = None
        self.cluster_children = None
        self.optimal_num_clusters = None
        self.returns_estimator = ReturnsEstimators()
        self.risk_estimator = RiskEstimators()
        self.risk_metrics = RiskMetrics()
        self.confidence_level = confidence_level

    def allocate(self, asset_names=None, asset_prices=None, asset_returns=None, covariance_matrix=None,
                 risk_measure='equal_weighting', linkage='ward', optimal_num_clusters=None):
        # pylint: disable=too-many-branches
        """
        Calculate asset allocations using the Hierarchical Equal Risk Contribution algorithm.

        :param asset_names: (list) A list of strings containing the asset names.
        :param asset_prices: (pd.DataFrame) A dataframe of historical asset prices (daily close)
                                            indexed by date.
        :param asset_returns: (pd.DataFrame/numpy matrix) User supplied matrix of asset returns.
        :param covariance_matrix: (pd.DataFrame/numpy matrix) User supplied covariance matrix of asset returns.
        :param risk_measure: (str) The metric used for calculating weight allocations. Supported strings - ``equal_weighting``,
                                   ``variance``, ``standard_deviation``, ``expected_shortfall``, ``conditional_drawdown_risk``.
        :param linkage: (str) The type of linkage method to use for clustering. Supported strings - ``single``, ``average``,
                              ``complete``, ``ward``.
        :param optimal_num_clusters: (int) Optimal number of clusters for hierarchical clustering.
        """

        # Perform error checks
        self._error_checks(asset_prices, asset_returns, risk_measure, covariance_matrix)

        if asset_names is None:
            if asset_prices is not None:
                asset_names = asset_prices.columns
            elif asset_returns is not None and isinstance(asset_returns, pd.DataFrame):
                asset_names = asset_returns.columns
            else:
                raise ValueError("Please provide a list of asset names")

        # Calculate the returns if the user does not supply a returns dataframe
        if asset_returns is None and (risk_measure in {'expected_shortfall', 'conditional_drawdown_risk'} or covariance_matrix is
                                      None or not optimal_num_clusters):
            asset_returns = self.returns_estimator.calculate_returns(asset_prices=asset_prices)
        asset_returns = pd.DataFrame(asset_returns, columns=asset_names)

        # Calculate covariance of returns or use the user specified covariance matrix
        if covariance_matrix is None:
            covariance_matrix = asset_returns.cov()
        cov = pd.DataFrame(covariance_matrix, index=asset_names, columns=asset_names)

        # Calculate correlation from covariance matrix
        corr = self.risk_estimator.cov_to_corr(cov)

        # Calculate the optimal number of clusters
        if not optimal_num_clusters:
            self.optimal_num_clusters = self._get_optimal_number_of_clusters(correlation=corr,
                                                                             linkage=linkage,
                                                                             asset_returns=asset_returns)
        else:
            self.optimal_num_clusters = self._check_max_number_of_clusters(num_clusters=optimal_num_clusters,
                                                                           linkage=linkage,
                                                                           correlation=corr)

        # Tree Clustering
        self.clusters, self.cluster_children = self._tree_clustering(correlation=corr,
                                                                     linkage=linkage)

        # Get the flattened order of assets in hierarchical clustering tree
        num_assets = len(asset_names)
        self.ordered_indices = self._quasi_diagnalization(num_assets, 2 * num_assets - 2)

        # Recursive Bisection
        self._recursive_bisection(asset_returns=asset_returns,
                                  covariance_matrix=cov,
                                  assets=asset_names,
                                  risk_measure=risk_measure)

    def plot_clusters(self, assets):
        """
        Plot a dendrogram of the hierarchical clusters.

        :param assets: (list) Asset names in the portfolio
        :return: (dict) Dendrogram
        """

        colors = dict()
        for cluster_idx, children in self.cluster_children.items():
            color = self.UniqueColors[cluster_idx]

            for child in children:
                colors[assets[child]] = color
        dendrogram_plot = dendrogram(self.clusters, labels=assets, link_color_func=lambda k: self.UnclusteredColor)
        plot_axis = plt.gca()
        xlbls = plot_axis.get_xmajorticklabels()
        for lbl in xlbls:
            lbl.set_color(colors[lbl.get_text()])
        return dendrogram_plot

    @staticmethod
    def _compute_cluster_inertia(labels, asset_returns):
        """
        Calculate the cluster inertia (within cluster sum-of-squares).

        :param labels: (list) Cluster labels.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :return: (float) Cluster inertia value.
        """

        unique_labels = np.unique(labels)
        inertia = [np.mean(pairwise_distances(asset_returns[:, labels == label])) for label in unique_labels]
        inertia = np.log(np.sum(inertia))
        return inertia

    @staticmethod
    def _check_max_number_of_clusters(num_clusters, linkage, correlation):
        """
        In some cases, the optimal number of clusters value given by the users is greater than the maximum number of clusters
        possible with the given data. This function checks this and assigns the proper value to the number of clusters when the
        given value exceeds maximum possible clusters.

        :param num_clusters: (int) The number of clusters.
        :param linkage (str): The type of linkage method to use for clustering.
        :param correlation: (np.array) Matrix of asset correlations.
        :return: (int) New value for number of clusters.
        """

        distance_matrix = np.sqrt(2 * (1 - correlation).round(5))
        clusters = scipy_linkage(squareform(distance_matrix.values), method=linkage)
        clustering_inds = fcluster(clusters, num_clusters, criterion='maxclust')
        max_number_of_clusters_possible = max(clustering_inds)
        num_clusters = min(max_number_of_clusters_possible, num_clusters)
        return num_clusters

    def _get_optimal_number_of_clusters(self, correlation, asset_returns, linkage, num_reference_datasets=5):
        """
        Find the optimal number of clusters for hierarchical clustering using the Gap statistic.

        :param correlation: (np.array) Matrix of asset correlations.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :param linkage: (str) The type of linkage method to use for clustering.
        :param num_reference_datasets: (int) The number of reference datasets to generate for calculating expected inertia.
        :return: (int) The optimal number of clusters.
        """

        original_distance_matrix = np.sqrt(2 * (1 - correlation).round(5))
        gap_values = []
        num_clusters = 1
        max_number_of_clusters = float("-inf")
        while True:

            # Calculate inertia from original data
            original_clusters = scipy_linkage(squareform(original_distance_matrix), method=linkage)
            original_cluster_assignments = fcluster(original_clusters, num_clusters, criterion='maxclust')
            if max(original_cluster_assignments) == max_number_of_clusters or max(original_cluster_assignments) > 10:
                break
            max_number_of_clusters = max(original_cluster_assignments)
            inertia = self._compute_cluster_inertia(original_cluster_assignments, asset_returns.values)

            # Calculate expected inertia from reference datasets
            expected_inertia = self._calculate_expected_inertia(num_reference_datasets, asset_returns, num_clusters, linkage)

            # Calculate the gap statistic
            gap = expected_inertia - inertia
            gap_values.append(gap)
            num_clusters += 1
        return 1 + np.argmax(gap_values)

    def _calculate_expected_inertia(self, num_reference_datasets, asset_returns, num_clusters, linkage):
        """
        Calculate the expected inertia by generating clusters from a uniform distribution.

        :param num_reference_datasets: (int) The number of reference datasets to generate from the distribution.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :param num_clusters: (int) The number of clusters to generate.
        :param linkage: (str) The type of linkage criterion to use for hierarchical clustering.
        :return: (float) The expected inertia from the reference datasets.
        """

        reference_inertias = []
        for _ in range(num_reference_datasets):
            # Generate reference returns from uniform distribution and calculate the distance matrix.
            reference_asset_returns = pd.DataFrame(np.random.rand(*asset_returns.shape))
            reference_correlation = np.array(reference_asset_returns.corr())
            reference_distance_matrix = np.sqrt(2 * (1 - reference_correlation).round(5))

            reference_clusters = scipy_linkage(squareform(reference_distance_matrix), method=linkage)
            reference_cluster_assignments = fcluster(reference_clusters, num_clusters, criterion='maxclust')
            inertia = self._compute_cluster_inertia(reference_cluster_assignments, reference_asset_returns.values)
            reference_inertias.append(inertia)
        return np.mean(reference_inertias)

    def _tree_clustering(self, correlation, linkage):
        """
        Perform agglomerative clustering on the current portfolio.

        :param correlation: (np.array) Matrix of asset correlations.
        :param linkage (str): The type of linkage method to use for clustering.
        :return: (list) Structure of hierarchical tree.
        """

        distance_matrix = np.sqrt(2 * (1 - correlation).round(5))
        clusters = scipy_linkage(squareform(distance_matrix.values), method=linkage)
        clustering_inds = fcluster(clusters, self.optimal_num_clusters, criterion='maxclust')
        cluster_children = {index - 1: [] for index in range(min(clustering_inds), max(clustering_inds) + 1)}
        for index, cluster_index in enumerate(clustering_inds):
            cluster_children[cluster_index - 1].append(index)
        return clusters, cluster_children

    def _quasi_diagnalization(self, num_assets, curr_index):
        """
        Rearrange the assets to reorder them according to hierarchical tree clustering order.

        :param num_assets: (int) The total number of assets.
        :param curr_index: (int) Current index.
        :return: (list) The assets rearranged according to hierarchical clustering.
        """

        if curr_index < num_assets:
            return [curr_index]

        left = int(self.clusters[curr_index - num_assets, 0])
        right = int(self.clusters[curr_index - num_assets, 1])

        return (self._quasi_diagnalization(num_assets, left) + self._quasi_diagnalization(num_assets, right))

    def _recursive_bisection(self, asset_returns, covariance_matrix, assets, risk_measure):
        """
        Recursively assign weights to the clusters - ultimately assigning weights to the individual assets.

        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :param covariance_matrix: (pd.DataFrame) The covariance matrix.
        :param assets: (list) List of asset names in the portfolio.
        :param risk_measure: (str) The metric used for calculating weight allocations.
        """

        num_assets = len(assets)
        self.weights = np.ones(shape=num_assets)
        clusters_contribution = np.ones(shape=self.optimal_num_clusters)
        clusters_weights = np.ones(shape=self.optimal_num_clusters)

        # Calculate the corresponding risk measure for the clusters
        self._calculate_risk_contribution_of_clusters(clusters_contribution,
                                                      risk_measure,
                                                      covariance_matrix,
                                                      asset_returns)

        # Recursive bisection taking into account the dendrogram structure
        for cluster_index in range(self.optimal_num_clusters - 1):

            # Get the left and right cluster ids
            left_cluster_ids, right_cluster_ids = self._get_children_cluster_ids(num_assets=num_assets,
                                                                                 parent_cluster_id=cluster_index)

            # Compute alpha
            left_cluster_contribution = np.sum(clusters_contribution[left_cluster_ids])
            right_cluster_contribution = np.sum(clusters_contribution[right_cluster_ids])
            if risk_measure == 'equal_weighting':
                alloc_factor = 0.5
            else:
                alloc_factor = 1 - left_cluster_contribution / (left_cluster_contribution + right_cluster_contribution)

            # Assign weights to each sub-cluster
            clusters_weights[left_cluster_ids] *= alloc_factor
            clusters_weights[right_cluster_ids] *= 1 - alloc_factor

        # Compute the final weights
        self._calculate_final_portfolio_weights(risk_measure,
                                                clusters_weights,
                                                covariance_matrix,
                                                asset_returns)

        # Assign actual asset names to weight index
        self.weights = pd.DataFrame(self.weights)
        self.weights.index = assets
        self.weights = self.weights.T
        self.weights = self.weights.iloc[:, self.ordered_indices]

    def _calculate_final_portfolio_weights(self, risk_measure, clusters_weights, covariance_matrix, asset_returns):
        """
        Calculate the final asset weights.

        :param risk_measure: (str) The metric used for calculating weight allocations.
        :param clusters_weights: (np.array) The cluster weights calculated using recursive bisection.
        :param covariance_matrix: (pd.DataFrame) The covariance matrix.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        """

        for cluster_index in range(self.optimal_num_clusters):
            cluster_asset_indices = self.cluster_children[cluster_index]

            # Covariance of assets in this cluster
            cluster_covariance = covariance_matrix.iloc[cluster_asset_indices, cluster_asset_indices]

            # Historical returns of assets in this cluster
            cluster_asset_returns = None
            if not asset_returns.empty:
                cluster_asset_returns = asset_returns.iloc[:, cluster_asset_indices]

            parity_weights = self._calculate_naive_risk_parity(cluster_index=cluster_index,
                                                               risk_measure=risk_measure,
                                                               covariance=cluster_covariance,
                                                               asset_returns=cluster_asset_returns)
            self.weights[cluster_asset_indices] = parity_weights * clusters_weights[cluster_index]

    def _calculate_naive_risk_parity(self, cluster_index, risk_measure, covariance, asset_returns):
        # pylint: disable=no-else-return
        """
        Calculate the naive risk parity weights.

        :param cluster_index: (int) Index of the current cluster.
        :param risk_measure: (str) The metric used for calculating weight allocations.
        :param covariance: (pd.DataFrame) The covariance matrix of asset returns.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :return: (np.array) list of risk parity weights for assets in current cluster.
        """

        if risk_measure == 'equal_weighting':
            num_assets_in_cluster = len(self.cluster_children[cluster_index])
            return np.ones(num_assets_in_cluster) * 1/num_assets_in_cluster
        elif risk_measure in {'variance', 'standard_deviation'}:
            return self._get_inverse_variance_weights(covariance)
        elif risk_measure == 'expected_shortfall':
            return self._get_inverse_CVaR_weights(asset_returns)
        return self._get_inverse_CDaR_weights(asset_returns)

    def _calculate_risk_contribution_of_clusters(self, clusters_contribution, risk_measure,
                                                 covariance_matrix, asset_returns):
        """
        Calculate the risk contribution of clusters based on the allocation metric.

        :param clusters_contribution: (np.array) The risk contribution value of the clusters.
        :param risk_measure: (str) The metric used for calculating weight allocations.
        :param covariance_matrix: (pd.DataFrame) The covariance matrix.
        :param asset_returns: (pd.DataFrame) Historical asset returns.
        """

        for cluster_index in range(self.optimal_num_clusters):
            cluster_asset_indices = self.cluster_children[cluster_index]

            if risk_measure == 'variance':
                clusters_contribution[cluster_index] = self._get_cluster_variance(covariance_matrix,
                                                                                  cluster_asset_indices)
            elif risk_measure == 'standard_deviation':
                clusters_contribution[cluster_index] = np.sqrt(
                    self._get_cluster_variance(covariance_matrix, cluster_asset_indices))
            elif risk_measure == 'expected_shortfall':
                clusters_contribution[cluster_index] = self._get_cluster_expected_shortfall(asset_returns,
                                                                                            cluster_asset_indices)
            elif risk_measure == 'conditional_drawdown_risk':
                clusters_contribution[cluster_index] = self._get_cluster_conditional_drawdown_at_risk(
                    asset_returns=asset_returns,
                    cluster_indices=cluster_asset_indices)

    def _get_children_cluster_ids(self, num_assets, parent_cluster_id):
        """
        Find the left and right children cluster id of the given parent cluster id.

        :param num_assets: (int) The number of assets in the portfolio.
        :param parent_cluster_index: (int) The current parent cluster id.
        :return: (list, list) List of cluster ids to the left and right of the parent cluster in the hierarchical tree.
        """

        left = int(self.clusters[num_assets - 2 - parent_cluster_id, 0])
        right = int(self.clusters[num_assets - 2 - parent_cluster_id, 1])
        left_cluster = self._quasi_diagnalization(num_assets, left)
        right_cluster = self._quasi_diagnalization(num_assets, right)

        left_cluster_ids = []
        right_cluster_ids = []
        for id_cluster, cluster in self.cluster_children.items():
            if sorted(self._intersection(left_cluster, cluster)) == sorted(cluster):
                left_cluster_ids.append(id_cluster)
            if sorted(self._intersection(right_cluster, cluster)) == sorted(cluster):
                right_cluster_ids.append(id_cluster)

        return left_cluster_ids, right_cluster_ids

    @staticmethod
    def _get_inverse_variance_weights(covariance):
        """
        Calculate inverse variance weight allocations.

        :param covariance: (pd.DataFrame) Covariance matrix of assets.
        :return: (np.array) Inverse variance weight values.
        """

        inv_diag = 1 / np.diag(covariance.values)
        parity_weights = inv_diag * (1 / np.sum(inv_diag))
        return parity_weights

    def _get_inverse_CVaR_weights(self, asset_returns):
        # pylint: disable=invalid-name
        """
        Calculate inverse CVaR weight allocations.

        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :return: (np.array) Inverse CVaR weight values.
        """

        parity_weights = []
        for asset_index in range(asset_returns.shape[1]):
            returns = asset_returns.iloc[:, asset_index]
            cvar = self.risk_metrics.calculate_expected_shortfall(returns=returns,
                                                                  confidence_level=self.confidence_level)
            parity_weights.append(cvar)

        parity_weights = np.array(parity_weights)
        parity_weights = 1 / parity_weights
        parity_weights = parity_weights * (1 / np.sum(parity_weights))
        return parity_weights

    def _get_inverse_CDaR_weights(self, asset_returns):
        # pylint: disable=invalid-name
        """
        Calculate inverse CDaR weight allocations.

        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :return: (np.array) Inverse CDaR weight values.
        """

        parity_weights = []
        for asset_index in range(asset_returns.shape[1]):
            returns = asset_returns.iloc[:, asset_index]
            cdar = self.risk_metrics.calculate_conditional_drawdown_risk(returns=returns,
                                                                         confidence_level=self.confidence_level)
            parity_weights.append(cdar)

        parity_weights = np.array(parity_weights)
        parity_weights = 1 / parity_weights
        parity_weights = parity_weights * (1 / np.sum(parity_weights))
        return parity_weights

    def _get_cluster_variance(self, covariance, cluster_indices):
        """
        Calculate cluster variance.

        :param covariance: (pd.DataFrame) Covariance matrix of asset returns.
        :param cluster_indices: (list) List of asset indices for the cluster.
        :return: (float) Variance of the cluster.
        """

        cluster_covariance = covariance.iloc[cluster_indices, cluster_indices]
        parity_weights = self._get_inverse_variance_weights(cluster_covariance)
        cluster_variance = self.risk_metrics.calculate_variance(covariance=cluster_covariance, weights=parity_weights)
        return cluster_variance

    def _get_cluster_expected_shortfall(self, asset_returns, cluster_indices):
        """
        Calculate cluster expected shortfall.

        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :param cluster_indices: (list) List of asset indices for the cluster.
        :return: (float) Expected shortfall of the cluster.
        """

        cluster_asset_returns = asset_returns.iloc[:, cluster_indices]
        parity_weights = self._get_inverse_CVaR_weights(cluster_asset_returns)
        portfolio_returns = cluster_asset_returns @ parity_weights
        cluster_expected_shortfall = self.risk_metrics.calculate_expected_shortfall(returns=portfolio_returns,
                                                                                    confidence_level=self.confidence_level)
        return cluster_expected_shortfall

    def _get_cluster_conditional_drawdown_at_risk(self, asset_returns, cluster_indices):
        """
        Calculate cluster conditional drawdown at risk.

        :param asset_returns: (pd.DataFrame) Historical asset returns.
        :param cluster_indices: (list) List of asset indices for the cluster.
        :return: (float) CDD of the cluster.
        """

        cluster_asset_returns = asset_returns.iloc[:, cluster_indices]
        parity_weights = self._get_inverse_CDaR_weights(cluster_asset_returns)
        portfolio_returns = cluster_asset_returns @ parity_weights
        cluster_conditional_drawdown = self.risk_metrics.calculate_conditional_drawdown_risk(returns=portfolio_returns,
                                                                                             confidence_level=self.confidence_level)
        return cluster_conditional_drawdown

    @staticmethod
    def _intersection(list1, list2):
        """
        Calculate the intersection of two lists

        :param list1: (list) The first list of items.
        :param list2: (list) The second list of items.
        :return: (list) List containing the intersection of the input lists.
        """

        return list(set(list1) & set(list2))

    @staticmethod
    def _error_checks(asset_prices, asset_returns, risk_measure, covariance_matrix):
        """
        Perform initial warning checks.

        :param asset_prices: (pd.DataFrame) A dataframe of historical asset prices (daily close)
                                            indexed by date.
        :param asset_returns: (pd.DataFrame/numpy matrix) User supplied matrix of asset returns.
        :param risk_measure: (str) The metric used for calculating weight allocations.
        :param covariance_matrix: (pd.DataFrame/numpy matrix) User supplied covariance matrix of asset returns.
        """

        if asset_prices is None and asset_returns is None and covariance_matrix is None:
            raise ValueError("You need to supply either raw prices or returns or covariance matrix")

        if asset_prices is not None:
            if not isinstance(asset_prices, pd.DataFrame):
                raise ValueError("Asset prices matrix must be a dataframe")
            if not isinstance(asset_prices.index, pd.DatetimeIndex):
                raise ValueError("Asset prices dataframe must be indexed by date.")

        if risk_measure not in \
                {'variance', 'standard_deviation', 'equal_weighting', 'expected_shortfall',
                 'conditional_drawdown_risk'}:
            raise ValueError("Unknown allocation metric specified. Supported metrics are - variance, "
                             "standard_deviation, equal_weighting, expected_shortfall, "
                             "conditional_drawdown_risk")



# pylint: disable=missing-module-docstring

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.labeling.return_vs_benchmark import return_over_benchmark


class TestReturnOverBenchmark(unittest.TestCase):
    """
    Tests regarding the labeling returns over benchmark method.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date', parse_dates=True)
        self.idx10 = self.data[:10].index

    def test_basic(self):
        """
        Tests for the basic case where the benchmark is a constant.
        """
        data1 = self.data['EWQ']
        data3 = self.data[['EWU', 'XLB']]

        # No benchmark given, assumed to be 0.
        test1 = return_over_benchmark(data1[:10], lag=True)
        test2 = return_over_benchmark(data1[:10], binary=True, lag=True)

        # Constant benchmark, on multiple columns.
        test3 = return_over_benchmark(data3[:10], benchmark=0.005, lag=False)

        test1_actual = pd.Series([-0.00052716, -0.02452523, 0.00729918, -0.00778307, 0.0029754, 0.00242709,
                                  -0.0191014, 0.02057049, -0.02983071, np.nan], index=self.idx10)
        test3_actual = pd.DataFrame({'EWU': [np.nan, 0.001281, -0.028720, -0.002016, -0.019025, -0.004569, -0.010170,
                                             -0.026221, 0.006947, -0.029924],
                                     'XLB': [np.nan, 0.013160, -0.035202, -0.018732, -0.020415, 0.001313, 0.008551,
                                             -0.015151, 0.028517, -0.032348]},
                                    index=self.idx10, columns=data3.columns)
        pd.testing.assert_series_equal(test1, test1_actual, check_names=False)
        pd.testing.assert_series_equal(test2, test1_actual.apply(np.sign), check_names=False)
        pd.testing.assert_frame_equal(test3, test3_actual, rtol=0.001)

    def test_given_benchmark(self):
        """
        Tests comparing value to a dynamic benchmark.
        """
        # User inputted benchmark.
        benchmark4 = pd.Series([0, 0.01, -0.01, 0.02, -0.005, 0.6, 100, -90, -0.2, 0.008], index=self.idx10)
        data4 = self.data['BND']
        test4 = return_over_benchmark(data4[:10], benchmark=benchmark4, lag=False)
        test5 = return_over_benchmark(data4[:10], benchmark=benchmark4, binary=True, lag=False)
        test4_actual = pd.Series([np.nan, -8.70736203e-03, 1.11619412e-02, -1.97421452e-02, 6.03135015e-03,
                                  -6.01159098e-01, -1.00000387e+02, 9.00030955e+01, 2.01285921e-01,
                                  -4.40427932e-03], index=self.idx10)

        # Using SPY as a benchmark.
        benchmark6 = self.data['SPY'].pct_change(periods=1)
        test6 = return_over_benchmark(data4[:10], benchmark6[:10], lag=False)
        test6_actual = pd.Series([np.nan, 0.00177558, 0.02566838, 0.00110702, 0.01717979, -0.01166944, -0.00694088,
                                  0.01116406, -0.0067769, 0.02560875], index=self.idx10)

        pd.testing.assert_series_equal(test4, test4_actual)
        pd.testing.assert_series_equal(test5, test4_actual.apply(np.sign))
        pd.testing.assert_series_equal(test6, test6_actual)

    def test_resample(self):
        """
        Tests for when resampling is used.
        """
        data5 = self.data[['EEM', 'EWG', 'TIP']]
        subset1 = data5[40:50]
        subset2 = data5[0:130]
        benchmark_day = pd.Series([0.01, 0.01, 0.01, -0.01, -0.01, -0.02, 0.2, 0.04, -0.1, 0], index=subset1.index)
        month_index = subset2.resample('M').last().index

        test7 = return_over_benchmark(subset1, benchmark=benchmark_day, binary=False, resample_by='B', lag=True)
        test7b = return_over_benchmark(subset1, benchmark=benchmark_day, binary=False, lag=True)
        test8 = return_over_benchmark(subset2, benchmark=-0.02, binary=True, resample_by='M', lag=True)  # Negative
        test8_actual = pd.DataFrame({'EEM': [1, -1, 1, 1, -1, -1, np.nan], 'EWG': [1, 1, 1, 1, -1, 1, np.nan],
                                     'TIP': [1, 1, -1, 1, 1, 1, np.nan]}, index=month_index)

        pd.testing.assert_frame_equal(test7, test7b)
        pd.testing.assert_frame_equal(test8, test8_actual)

    def test_exception(self):
        """
        Verifies that the exception is given when there is a mismatch between prices.index and benchmark.index.
        """
        returns = self.data['TLT'].pct_change()
        benchmark = self.data['SPY'].pct_change()

        # Suppose we resample the returns, but fail to update the index of benchmark.
        with self.assertRaises(Exception):
            return_over_benchmark(returns, benchmark=benchmark, resample_by='W')



"""
Test various functions regarding chapter 18: Microstructural Features.
"""

import os
import unittest

import numpy as np
import pandas as pd

from famlafl.data_structures import get_volume_bars
from famlafl.microstructural_features import (get_vpin, get_bar_based_amihud_lambda, get_bar_based_kyle_lambda,
                                               get_bekker_parkinson_vol, get_corwin_schultz_estimator,
                                               get_bar_based_hasbrouck_lambda, get_roll_impact, get_roll_measure,
                                               quantile_mapping, sigma_mapping, MicrostructuralFeaturesGenerator)
from famlafl.microstructural_features.encoding import encode_tick_rule_array
from famlafl.microstructural_features.entropy import get_plug_in_entropy, get_shannon_entropy, get_lempel_ziv_entropy, \
    get_konto_entropy, _match_length
from famlafl.util import get_bvc_buy_volume


class TestMicrostructuralFeatures(unittest.TestCase):
    """
    Test get_inter_bar_features, test_first_generation, test_second_generation, test_misc
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.trades_path = project_path + '/test_data/tick_data.csv'
        self.data = pd.read_csv(self.path, index_col='date_time', parse_dates=[0])
        self.data.index = pd.to_datetime(self.data.index)

    def test_first_generation(self):
        """
        Test first generation intra-bar features
        """

        roll_measure = get_roll_measure(self.data.close, window=20)
        roll_impact = get_roll_impact(self.data.close, self.data.cum_dollar, window=20)
        corwin_schultz = get_corwin_schultz_estimator(self.data.high, self.data.low, window=20)
        bekker_parkinson = get_bekker_parkinson_vol(self.data.high, self.data.low, window=20)

        # Shape assertions
        self.assertEqual(self.data.shape[0], roll_measure.shape[0])
        self.assertEqual(self.data.shape[0], roll_impact.shape[0])
        self.assertEqual(self.data.shape[0], corwin_schultz.shape[0])
        self.assertEqual(self.data.shape[0], bekker_parkinson.shape[0])

        # Roll measure/impact assertions
        self.assertAlmostEqual(roll_measure.max(), 7.1584, delta=1e-4)
        self.assertAlmostEqual(roll_measure.mean(), 2.341, delta=1e-3)
        self.assertAlmostEqual(roll_measure.iloc[25], 1.176, delta=1e-3)  # Test specific position using iloc

        self.assertAlmostEqual(roll_impact.max(), 1.022e-7, delta=1e-7)
        self.assertAlmostEqual(roll_impact.mean(), 3.3445e-8, delta=1e-7)
        self.assertAlmostEqual(roll_impact.iloc[25], 1.6807e-8, delta=1e-4)  # Use iloc for positional indexing

        # Test Corwin-Schultz
        self.assertAlmostEqual(corwin_schultz.max(), 0.01652, delta=1e-4)
        self.assertAlmostEqual(corwin_schultz.mean(), 0.00151602, delta=1e-4)
        self.assertAlmostEqual(corwin_schultz.iloc[25], 0.00139617, delta=1e-4)  # Use iloc for positional indexing

        self.assertAlmostEqual(bekker_parkinson.max(), 0.018773, delta=1e-4)
        self.assertAlmostEqual(bekker_parkinson.mean(), 0.001456, delta=1e-4)
        self.assertAlmostEqual(bekker_parkinson.iloc[25], 0.000517, delta=1e-4)  # Use iloc for positional indexing

    def test_second_generation_intra_bar(self):
        """
        Test intra-bar second generation features
        """
        kyle_lambda = get_bar_based_kyle_lambda(self.data.close, self.data.cum_vol, window=20)
        amihud_lambda = get_bar_based_amihud_lambda(self.data.close, self.data.cum_dollar, window=20)
        hasbrouck_lambda = get_bar_based_hasbrouck_lambda(self.data.close, self.data.cum_dollar, window=20)

        # Shape assertions
        self.assertEqual(self.data.shape[0], kyle_lambda.shape[0])
        self.assertEqual(self.data.shape[0], amihud_lambda.shape[0])
        self.assertEqual(self.data.shape[0], hasbrouck_lambda.shape[0])

        # Test Kyle Lambda
        self.assertAlmostEqual(kyle_lambda.max(), 0.000163423, delta=1e-6)
        self.assertAlmostEqual(kyle_lambda.mean(), 7.02e-5, delta=1e-6)
        self.assertAlmostEqual(kyle_lambda.iloc[25], 7.76e-5, delta=1e-6)  # Test specific position using iloc

        # Test Amihud Lambda
        self.assertAlmostEqual(amihud_lambda.max(), 4.057838e-11, delta=1e-13)
        self.assertAlmostEqual(amihud_lambda.mean(), 1.7213e-11, delta=1e-13)
        self.assertAlmostEqual(amihud_lambda.iloc[25], 1.8439e-11, delta=1e-13)  # Use iloc for positional indexing

        # Test Hasbrouck lambda
        self.assertAlmostEqual(hasbrouck_lambda.max(), 3.39527e-7, delta=1e-10)
        self.assertAlmostEqual(hasbrouck_lambda.mean(), 1.44037e-7, delta=1e-10)
        self.assertAlmostEqual(hasbrouck_lambda.iloc[25], 1.5433e-7, delta=1e-10)  # Use iloc for positional indexing

    def test_third_generation(self):
        """
        Test third generation features
        """
        bvc_buy_volume = get_bvc_buy_volume(self.data.close, self.data.cum_vol, window=20)
        vpin_1 = get_vpin(self.data.cum_vol, bvc_buy_volume)
        vpin_20 = get_vpin(self.data.cum_vol, bvc_buy_volume, window=20)

        self.assertEqual(self.data.shape[0], vpin_1.shape[0])
        self.assertEqual(self.data.shape[0], vpin_20.shape[0])

        self.assertAlmostEqual(vpin_1.max(), 0.999, delta=1e-3)
        self.assertAlmostEqual(vpin_1.mean(), 0.501, delta=1e-3)
        self.assertAlmostEqual(vpin_1.iloc[25], 0.554, delta=1e-3)  # Use iloc for positional indexing

        self.assertAlmostEqual(vpin_20.max(), 0.6811, delta=1e-3)
        self.assertAlmostEqual(vpin_20.mean(), 0.500, delta=1e-3)
        self.assertAlmostEqual(vpin_20.iloc[45], 0.4638, delta=1e-3)  # Use iloc for positional indexing

    def test_tick_rule_encoding(self):
        """
        Test tick rule encoding function
        """
        with self.assertRaises(ValueError):
            encode_tick_rule_array([-1, 1, 0, 20000000])

        encoded_tick_rule = encode_tick_rule_array([-1, 1, 0, 0])
        self.assertEqual('bacc', encoded_tick_rule)

    def test_entropy_calculations(self):
        """
        Test entropy functions
        """
        message = '11100001'
        message_array = [1, 1, 1, 0, 0, 0, 0, 1]
        shannon = get_shannon_entropy(message)
        plug_in = get_plug_in_entropy(message, word_length=1)
        plug_in_arr = get_plug_in_entropy(message_array, word_length=1)
        lempel = get_lempel_ziv_entropy(message)
        konto = get_konto_entropy(message)

        self.assertEqual(plug_in, plug_in_arr)
        self.assertAlmostEqual(shannon, 1.0, delta=1e-3)
        self.assertAlmostEqual(lempel, 0.625, delta=1e-3)
        self.assertAlmostEqual(plug_in, 0.985, delta=1e-3)
        self.assertAlmostEqual(konto, 0.9682, delta=1e-3)

        # Konto entropy boundary conditions
        konto_2 = get_konto_entropy(message, 2)
        _match_length('1101111', 2, 3)
        self.assertAlmostEqual(konto_2, 0.8453, delta=1e-4)
        self.assertEqual(get_konto_entropy('a'), 0)  # one-character message entropy = 0

    def test_encoding_schemes(self):
        """
        Test quantile and sigma encoding
        """
        values = np.arange(0, 1000, 1)
        quantile_dict = quantile_mapping(values, num_letters=10)
        sigma_dict = sigma_mapping(values, step=20)
        self.assertEqual(len(quantile_dict), 10)
        self.assertEqual(quantile_dict[229.77], '\x02')
        self.assertEqual(len(sigma_dict), np.ceil((max(values) - min(values)) / 20))
        self.assertEqual(sigma_dict[100], '\x05')

        with self.assertRaises(ValueError):
            sigma_mapping(values, step=1)  # Length of dice > ASCII table

    def test_csv_format(self):
        """
        Asserts that the csv data being passed is of the correct format.
        Tests various invalid input scenarios including:
        - Invalid date formats
        - Wrong data types for price and volume
        - Incorrect number of columns
        - Empty dataframes
        - Single column dataframes
        """
        # Test invalid date formats
        wrong_date = ['2019-41-30', 200.00, 5]  # Invalid month
        invalid_date = ['not-a-date', 200.00, 5]  # Non-date string
        wrong_price = ['2019-01-30', 'asd', 5]  # Non-float price
        wrong_volume = ['2019-01-30', 200.00, '1.5']  # String volume
        too_many_cols = ['2019-01-30', 200.00, 5, 'Limit order', 'B23']  # Extra columns
        empty_df = pd.DataFrame()  # Empty dataframe
        single_col = pd.DataFrame(['2019-01-30'])  # Single column dataframe

        # Test various invalid input scenarios
        with self.assertRaises(ValueError):
            MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(wrong_date).T)
        
        with self.assertRaises(ValueError):
            MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(invalid_date).T)

        with self.assertRaises(AssertionError):
            MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(too_many_cols).T)

        with self.assertRaises(AssertionError):
            MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(wrong_price).T)

        with self.assertRaises(AssertionError):
            MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(wrong_volume).T)

        with self.assertRaises(AssertionError):
            MicrostructuralFeaturesGenerator._assert_csv(empty_df)

        with self.assertRaises(AssertionError):
            MicrostructuralFeaturesGenerator._assert_csv(single_col)

        # Test valid input
        valid_data = ['2019-01-30', 200.00, 5]
        MicrostructuralFeaturesGenerator._assert_csv(pd.DataFrame(valid_data).T)

    def test_feature_generator_function(self):
        """
        Test validity of MicrostructuralFeaturesGenerator
        """
        # Encode volumes and pct changes
        df_trades = pd.read_csv(self.trades_path, parse_dates=[0])
        df_trades['log_ret'] = np.log(df_trades.Price / df_trades.Price.shift(1)).dropna()
        non_null_log_ret = df_trades[df_trades.log_ret != 0].log_ret.dropna()

        # Take unique volumes only
        volume_mapping = quantile_mapping(df_trades.Volume.drop_duplicates(), num_letters=10)

        returns_mapping = quantile_mapping(non_null_log_ret, num_letters=10)

        # Compress bars from ticks
        compressed_bars = get_volume_bars(self.trades_path, threshold=20, verbose=False)
        compressed_bars.set_index('date_time', inplace=True)
        compressed_bars.index = pd.to_datetime(compressed_bars.index)
        bar_index = compressed_bars.index

        # Test None input ValureError raise
        with self.assertRaises(ValueError):
            MicrostructuralFeaturesGenerator(None, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                             pct_encoding=returns_mapping)

        gen = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                               pct_encoding=returns_mapping)
        gen_no_entropy = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=None,
                                                          pct_encoding=None)
        gen_csv = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                                   pct_encoding=returns_mapping)
        gen_1 = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                                 pct_encoding=returns_mapping, batch_size=1)
        gen_20 = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                                  pct_encoding=returns_mapping, batch_size=20)
        gen_df = MicrostructuralFeaturesGenerator(df_trades, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                                  pct_encoding=returns_mapping, batch_size=20)
        features = gen.get_features(to_csv=False, verbose=False)
        features_1 = gen_1.get_features(to_csv=False, verbose=False)
        features_20 = gen_20.get_features(to_csv=False, verbose=False)
        features_from_df = gen_df.get_features(to_csv=False, verbose=False)
        features_no_entropy = gen_no_entropy.get_features(verbose=False)

        # No volume/pct entropy columns check
        with self.assertRaises(KeyError):
            features['tick_rule_entropy'] += features_no_entropy['volume_entropy_plug_in']

        with self.assertRaises(KeyError):
            features['tick_rule_entropy'] = features_no_entropy['pct_entropy_plug_in']

        gen_csv.get_features(to_csv=True, output_path='features.csv')
        features_from_csv = pd.read_csv('features.csv', parse_dates=[0])

        self.assertTrue((features.dropna().values == features_1.dropna().values).all())
        self.assertTrue((features.dropna().values == features_20.dropna().values).all())
        self.assertTrue((features.dropna().values == features_from_df.dropna().values).all())

        features.set_index('date_time', inplace=True)
        features_from_csv.set_index('date_time', inplace=True)

        self.assertAlmostEqual((features - features_from_csv).sum().sum(), 0, delta=1e-6)

        self.assertEqual(bar_index.shape[0], features.shape[0])
        self.assertEqual(compressed_bars.loc[features.index].shape[0], compressed_bars.shape[0])

        os.remove('features.csv')

    def test_inter_bar_feature_values(self):
        """
        Test entropy, misc, inter-bar feature generation
        """
        # Encode volumes and pct changes
        df_trades = pd.read_csv(self.trades_path, parse_dates=[0])
        df_trades['log_ret'] = np.log(df_trades.Price / df_trades.Price.shift(1)).dropna()
        unique_volumes = df_trades.Volume.drop_duplicates()
        non_null_log_ret = df_trades[df_trades.log_ret != 0].log_ret.dropna()

        volume_mapping = quantile_mapping(unique_volumes, num_letters=10)
        returns_mapping = quantile_mapping(non_null_log_ret, num_letters=10)

        # Compress bars from ticks
        compressed_bars = get_volume_bars(self.trades_path, threshold=20, verbose=False)
        compressed_bars.set_index('date_time', inplace=True)
        compressed_bars.index = pd.to_datetime(compressed_bars.index)

        gen = MicrostructuralFeaturesGenerator(self.trades_path, compressed_bars.tick_num, volume_encoding=volume_mapping,
                                               pct_encoding=returns_mapping)

        features = gen.get_features(to_csv=False, verbose=False)
        features.set_index('date_time', inplace=True)

        # Check individual feature values
        # Avg tick size
        self.assertAlmostEqual(features.avg_tick_size.max(), 8.0, delta=1e-1)
        self.assertAlmostEqual(features.avg_tick_size.mean(), 3.1931, delta=1e-4)
        self.assertAlmostEqual(features.avg_tick_size.iloc[3], 1.6153, delta=1e-3)  # Use iloc for positional indexing

        # Tick rule sum
        self.assertAlmostEqual(features.tick_rule_sum.max(), 7.0, delta=1e-1)
        self.assertAlmostEqual(features.tick_rule_sum.mean(), -3.4, delta=1e-4)
        self.assertAlmostEqual(features.tick_rule_sum.iloc[3], -11.0, delta=1e-3)  # Use iloc for positional indexing

        # VWAP
        self.assertAlmostEqual(features.vwap.max(), 1311.663, delta=1e-1)
        self.assertAlmostEqual(features.vwap.mean(), 1304.94542, delta=1e-4)
        self.assertAlmostEqual(features.vwap.iloc[3], 1304.5119, delta=1e-3)  # Use iloc for positional indexing

        # Kyle lambda
        self.assertAlmostEqual(features.kyle_lambda.max(), 197.958, delta=1e-1)
        self.assertAlmostEqual(features.kyle_lambda.mean(), 23.13859, delta=1e-4)
        self.assertAlmostEqual(features.kyle_lambda.iloc[3], 0.007936, delta=1e-3)  # Use iloc for positional indexing

        # Amihud lambda
        self.assertAlmostEqual(features.amihud_lambda.max(), 8.291e-5, delta=1e-7)
        self.assertAlmostEqual(features.amihud_lambda.mean(), 1.001e-5, delta=1e-8)
        self.assertAlmostEqual(features.amihud_lambda.iloc[3], 4.663786e-9, delta=1e-11)  # Use iloc for positional indexing

        # Hasbrouck lambda
        self.assertAlmostEqual(features.hasbrouck_lambda.max(), 0.0025621, delta=1e-5)
        self.assertAlmostEqual(features.hasbrouck_lambda.mean(), 0.00018253, delta=1e-5)
        self.assertAlmostEqual(features.hasbrouck_lambda.iloc[3], 2.42e-11, delta=1e-13)  # Use iloc for positional indexing

        # Tick rule entropy shannon
        self.assertAlmostEqual(features.tick_rule_entropy_shannon.max(), 1.52192, delta=1e-4)
        self.assertAlmostEqual(features.tick_rule_entropy_shannon.mean(), 0.499, delta=1e-4)
        self.assertAlmostEqual(features.tick_rule_entropy_shannon.iloc[3], 0.39124, delta=1e-4)  # Use iloc for positional indexing

        # Volume entropy plug-in
        self.assertAlmostEqual(features.volume_entropy_plug_in.max(), 1.92192, delta=1e-4)
        self.assertAlmostEqual(features.volume_entropy_plug_in.mean(), 1.052201, delta=1e-5)
        self.assertAlmostEqual(features.volume_entropy_plug_in.iloc[3], 0.41381, delta=1e-4)  # Use iloc for positional indexing

        # Volume entropy Lempel-Ziv
        self.assertAlmostEqual(features.volume_entropy_lempel_ziv.max(), 1.0, delta=1e-4)
        self.assertAlmostEqual(features.volume_entropy_lempel_ziv.mean(), 0.5904612, delta=1e-4)
        self.assertAlmostEqual(features.volume_entropy_lempel_ziv.iloc[3], 0.46153, delta=1e-4)  # Use iloc for positional indexing

        # Pct entropy Lempel-Ziv
        self.assertAlmostEqual(features.pct_entropy_lempel_ziv.max(), 0.8, delta=1e-4)
        self.assertAlmostEqual(features.pct_entropy_lempel_ziv.mean(), 0.56194, delta=1e-5)
        self.assertAlmostEqual(features.pct_entropy_lempel_ziv.iloc[3], 0.46153, delta=1e-5)  # Use iloc for positional indexing

        # Pct entropy Konto
        self.assertAlmostEqual(features.pct_entropy_konto.max(), 1.361, delta=1e-4)
        self.assertAlmostEqual(features.pct_entropy_konto.mean(), 0.83039791, delta=1e-5)
        self.assertAlmostEqual(features.pct_entropy_konto.iloc[3], 1.067022, delta=1e-5)  # Use iloc for positional indexing



"""
Tests Universal Portfolio.
"""

from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.up import UP


class TestUniversalPortfolio(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the UP class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_up_solution(self):
        """
        Test the calculation of UP weights.
        """
        # Initialize UP.
        up1 = UP(2)
        # Allocates asset prices to UP.
        up1.allocate(self.data)
        # Create np.array of all_weights.
        all_weights = np.array(up1.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_up_progress_solution(self):
        """
        Tests that UP prints progress bar.
        """
        # Initialize UP.
        up2 = UP(2)
        # Allocates asset prices to UP.
        up2.allocate(self.data, verbose=True)

    def test_up_uniform_solution(self):
        """
        Tests UP with uniform capital allocation.
        """
        # Initialize UP.
        up3 = UP(2, weighted='uniform')
        # Allocates asset prices to UP.
        up3.allocate(self.data)
        # Create np.array of all_weights.
        all_weights = np.array(up3.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_up_top_k_solution(self):
        """
        Tests UP with top-k experts capital allocation.
        """
        # Initialize UP.
        up4 = UP(5, weighted='top-k', k=2)
        # Allocates asset prices to UP.
        up4.allocate(self.data)
        # Create np.array of all_weights.
        all_weights = np.array(up4.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_up_wrong_method(self):
        """
        Tests ValueError if the method is not 'hist_performance', 'uniform', or 'top-k'.
        """
        # Initialize UP.
        up5 = UP(5, weighted='random', k=2)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            up5.allocate(self.data)

    def test_up_recalculate_solution(self):
        """
        Tests recalculate method in UP.
        """
        # Initialize UP.
        up6 = UP(3, weighted='top-k', k=2)
        # Allocates asset prices to UP.
        up6.allocate(self.data)
        # Recalculate with k=1.
        up6.recalculate_k(1)
        # Create np.array of all_weights.
        all_weights = np.array(up6.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_up_recalculate_error(self):
        """
        Tests ValueError if k is greater number of experts for recalculate.
        """
        # Initialize UP.
        up7 = UP(3, weighted='top-k', k=2)
        # Allocates asset prices to UP.
        up7.allocate(self.data)
        with self.assertRaises(ValueError):
            # Recalculate will raise ValueError.
            up7.recalculate_k(4)

    def test_up_recalculate1_error(self):
        """
        Tests ValueError if k is not an integer for recalculate.
        """
        # Initialize UP.
        up8 = UP(3, weighted='top-k', k=2)
        # Allocates asset prices to UP.
        up8.allocate(self.data)
        with self.assertRaises(ValueError):
            # Recalculate will raise ValueError.
            up8.recalculate_k(1.5)

    def test_up_recalculate2_error(self):
        """
        Tests ValueError if k is not greater than or equal to 1.
        """
        # Initialize UP.
        up9 = UP(3, weighted='top-k', k=2)
        # Allocates asset prices to UP.
        up9.allocate(self.data)
        with self.assertRaises(ValueError):
            # Recalculate will raise ValueError.
            up9.recalculate_k(0)



"""
Logic regarding sequential bootstrapping from chapter 4.
"""

import pandas as pd
import numpy as np
from numba import jit, prange
from numpy import float64


def get_ind_matrix(samples_info_sets, price_bars):
    """
    Advances in Financial Machine Learning, Snippet 4.3, page 65.

    Build an Indicator Matrix

    Get indicator matrix. The book implementation uses bar_index as input, however there is no explanation
    how to form it. We decided that using triple_barrier_events and price bars by analogy with concurrency
    is the best option.

    :param samples_info_sets: (pd.Series): Triple barrier events(t1) from labeling.get_events
    :param price_bars: (pd.DataFrame): Price bars which were used to form triple barrier events
    :return: (np.array) Indicator binary matrix indicating what (price) bars influence the label for each observation
    """
    if bool(samples_info_sets.isnull().values.any()) is True or bool(
            samples_info_sets.index.isnull().any()) is True:
        raise ValueError('NaN values in triple_barrier_events, delete nans')

    triple_barrier_events = pd.DataFrame(samples_info_sets)  # Convert Series to DataFrame

    # Take only period covered in triple_barrier_events
    trimmed_price_bars_index = price_bars[(price_bars.index >= triple_barrier_events.index.min()) &
                                          (price_bars.index <= triple_barrier_events.t1.max())].index

    label_endtime = triple_barrier_events.t1
    bar_index = list(triple_barrier_events.index)  # Generate index for indicator matrix from t1 and index
    bar_index.extend(triple_barrier_events.t1)
    bar_index.extend(trimmed_price_bars_index)  # Add price bars index
    bar_index = sorted(list(set(bar_index)))  # Drop duplicates and sort

    # Get sorted timestamps with index in sorted array
    sorted_timestamps = dict(zip(bar_index, range(len(bar_index))))

    # Map indices and end times to sorted timestamps
    label_start_positions = label_endtime.index.map(sorted_timestamps).astype(int)
    label_end_positions = label_endtime.map(sorted_timestamps).astype(int)


    tokenized_endtimes = np.column_stack((label_start_positions, label_end_positions))
    
    ind_mat = np.zeros((len(bar_index), len(label_endtime)))  # Init indicator matrix
    for sample_num, label_array in enumerate(tokenized_endtimes):
        label_index = label_array[0]
        label_endtime = label_array[1]
        ones_array = np.ones(
            (1, label_endtime - label_index + 1))  # Ones array which corresponds to number of 1 to insert
        ind_mat[label_index:label_endtime + 1, sample_num] = ones_array
    return ind_mat


def get_ind_mat_average_uniqueness(ind_mat):
    """
    Advances in Financial Machine Learning, Snippet 4.4. page 65.

    Compute Average Uniqueness

    Average uniqueness from indicator matrix

    :param ind_mat: (np.matrix) Indicator binary matrix
    :return: (float) Average uniqueness
    """
    if ind_mat.size == 0:
        return 0.0
    ind_mat = np.array(ind_mat, dtype=float64)
    concurrency = ind_mat.sum(axis=1)
    uniqueness = np.divide(ind_mat.T, concurrency, out=np.zeros_like(ind_mat.T), where=concurrency != 0)

    avg_uniqueness = uniqueness[uniqueness > 0].mean()

    return avg_uniqueness


def get_ind_mat_label_uniqueness(ind_mat):
    """
    Advances in Financial Machine Learning, An adaption of Snippet 4.4. page 65.

    Returns the indicator matrix element uniqueness.

    :param ind_mat: (np.matrix) Indicator binary matrix
    :return: (np.matrix) Element uniqueness
    """
    if ind_mat.size == 0:
        return np.array([])
    ind_mat = np.array(ind_mat, dtype=float64)
    concurrency = ind_mat.sum(axis=1)
    uniqueness = np.divide(ind_mat.T, concurrency, out=np.zeros_like(ind_mat.T), where=concurrency != 0)
    return uniqueness


@jit(parallel=True, nopython=True)
def _bootstrap_loop_run(ind_mat, prev_concurrency):  # pragma: no cover
    """
    Part of Sequential Bootstrapping for-loop. Using previously accumulated concurrency array, loops through all samples
    and generates averages uniqueness array of label based on previously accumulated concurrency

    :param ind_mat (np.array): Indicator matrix from get_ind_matrix function
    :param prev_concurrency (np.array): Accumulated concurrency from previous iterations of sequential bootstrapping
    :return: (np.array): Label average uniqueness based on prev_concurrency
    """
    avg_unique = np.zeros(ind_mat.shape[1])  # Array of label uniqueness

    for i in prange(ind_mat.shape[1]):  # pylint: disable=not-an-iterable
        prev_average_uniqueness = 0
        number_of_elements = 0
        reduced_mat = ind_mat[:, i]
        for j in range(len(reduced_mat)):  # pylint: disable=consider-using-enumerate
            if reduced_mat[j] > 0:
                new_el = reduced_mat[j] / (reduced_mat[j] + prev_concurrency[j])
                average_uniqueness = (prev_average_uniqueness * number_of_elements + new_el) / (number_of_elements + 1)
                number_of_elements += 1
                prev_average_uniqueness = average_uniqueness
        avg_unique[i] = average_uniqueness
    return avg_unique


def seq_bootstrap(ind_mat, sample_length=None, warmup_samples=None, compare=False, verbose=False,
                  random_state=np.random.RandomState()):
    """
    Advances in Financial Machine Learning, Snippet 4.5, Snippet 4.6, page 65.

    Return Sample from Sequential Bootstrap

    Generate a sample via sequential bootstrap.
    Note: Moved from pd.DataFrame to np.matrix for performance increase

    :param ind_mat: (pd.DataFrame) Indicator matrix from triple barrier events
    :param sample_length: (int) Length of bootstrapped sample
    :param warmup_samples: (list) List of previously drawn samples
    :param compare: (boolean) Flag to print standard bootstrap uniqueness vs sequential bootstrap uniqueness
    :param verbose: (boolean) Flag to print updated probabilities on each step
    :param random_state: (np.random.RandomState) Random state
    :return: (array) Bootstrapped samples indexes
    """

    if sample_length is None:
        sample_length = ind_mat.shape[1]

    if warmup_samples is None:
        warmup_samples = []

    phi = []  # Bootstrapped samples
    prev_concurrency = np.zeros(ind_mat.shape[0])  # Init with zeros (phi is empty)
    while len(phi) < sample_length:
        avg_unique = _bootstrap_loop_run(ind_mat, prev_concurrency)
        prob = avg_unique / sum(avg_unique)  # Draw prob
        try:
            choice = warmup_samples.pop(0)  # It would get samples from warmup until it is empty
            # If it is empty from the beginning it would get samples based on prob from the first iteration
        except IndexError:
            choice = random_state.choice(range(ind_mat.shape[1]), p=prob)
        phi += [choice]
        prev_concurrency += ind_mat[:, choice]  # Add recorded label array from ind_mat
        if verbose is True:
            print(prob)

    if compare is True:
        standard_indx = np.random.choice(ind_mat.shape[1], size=sample_length)
        standard_unq = get_ind_mat_average_uniqueness(ind_mat[:, standard_indx])
        sequential_unq = get_ind_mat_average_uniqueness(ind_mat[:, phi])
        print('Standard uniqueness: {}\nSequential uniqueness: {}'.format(standard_unq, sequential_unq))

    return phi



# pylint: disable=missing-module-docstring
import numpy as np
import scipy.optimize as opt
from famlafl.online_portfolio_selection.base import OLPS


class FTL(OLPS):
    """
    This class implements the Follow the Leader strategy. It is reproduced with modification from
    the following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    Follow the Leader strategy directly tracks the Best Constant Rebalanced Portfolio until the
    previous period.
    """

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight by following the Best Constant Rebalanced
        Portfolio.

        :param time: (int) Current time period.
        :return: (np.array) New portfolio weights by calculating the BCRP until the
                                        last time period.
        """
        # For time 0, return initial weights.
        if time == 0:
            return self.weights
        # Calculate BCRP weights until the current window.
        new_weights = self._fast_optimize(self.relative_return[:time + 1])
        return new_weights

    def _fast_optimize(self, optimize_array):
        """
        Calculates weights that maximize returns over the given array.

        :param optimize_array: (np.array) Relative returns of the assets for a given time period.
        :return: (np.array) Weights that maximize the returns for the given array.
        """
        # Initialize guess.
        weights = self._uniform_weight()

        # Use np.log and np.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        def _objective(weight):
            return -np.sum(np.log(np.dot(optimize_array, weight)))

        # Weight bounds.
        bounds = tuple((0.0, 1.0) for asset in range(self.number_of_assets))

        # Sum of weights is 1.
        const = ({'type': 'eq', 'fun': lambda w: np.sum(w) - 1})

        problem = opt.minimize(_objective, weights, method='SLSQP', bounds=bounds, constraints=const)
        return problem.x



"""
Implementation of Sequentially Bootstrapped Bagging Classifier using sklearn's library as base class
"""

from famlafl.ensemble.sb_bagging import SequentiallyBootstrappedBaggingClassifier, \
                                         SequentiallyBootstrappedBaggingRegressor



"""
Structural breaks test (CUSUM, Chow, SADF)
"""
from famlafl.structural_breaks.chow import get_chow_type_stat
from famlafl.structural_breaks.cusum import get_chu_stinchcombe_white_statistics
from famlafl.structural_breaks.sadf import get_sadf



"""
Contains the logic regarding the sequential bootstrapping from chapter 4, as well as the concurrent labels.
"""

from famlafl.sampling.bootstrapping import (get_ind_matrix, get_ind_mat_average_uniqueness, seq_bootstrap,
                                             get_ind_mat_label_uniqueness)
from famlafl.sampling.concurrent import (num_concurrent_events, _get_average_uniqueness,
                                          get_av_uniqueness_from_triple_barrier)



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd


class RiskMetrics:
    """
    This class contains methods for calculating common risk metrics used in trading and asset management.
    """

    def __init__(self):
        return

    @staticmethod
    def calculate_variance(covariance, weights):
        """
        Calculate the variance of a portfolio.

        :param covariance: (pd.DataFrame/np.matrix) Covariance matrix of assets
        :param weights: (list) List of asset weights
        :return: (float) Variance of a portfolio
        """

        return np.dot(weights, np.dot(covariance, weights))

    @staticmethod
    def calculate_value_at_risk(returns, confidence_level=0.05):
        """
        Calculate the value at risk (VaR) of a portfolio/asset.

        :param returns: (pd.DataFrame/np.array) Historical returns for an asset / portfolio
        :param confidence_level: (float) Confidence level (alpha)
        :return: (float) VaR
        """

        if not isinstance(returns, pd.DataFrame):
            returns = pd.DataFrame(returns)

        return returns.quantile(confidence_level, interpolation='higher')[0]

    def calculate_expected_shortfall(self, returns, confidence_level=0.05):
        """
        Calculate the expected shortfall (CVaR) of a portfolio/asset.

        :param returns: (pd.DataFrame/np.array) Historical returns for an asset / portfolio
        :param confidence_level: (float) Confidence level (alpha)
        :return: (float) Expected shortfall
        """

        if not isinstance(returns, pd.DataFrame):
            returns = pd.DataFrame(returns)

        value_at_risk = self.calculate_value_at_risk(returns, confidence_level)
        expected_shortfall = np.nanmean(returns[returns < value_at_risk])
        return expected_shortfall

    @staticmethod
    def calculate_conditional_drawdown_risk(returns, confidence_level=0.05):
        """
        Calculate the conditional drawdown of risk (CDaR) of a portfolio/asset.

        :param returns: (pd.DataFrame/np.array) Historical returns for an asset / portfolio
        :param confidence_level: (float) Confidence level (alpha)
        :return: (float) Conditional drawdown risk
        """

        if not isinstance(returns, pd.DataFrame):
            returns = pd.DataFrame(returns)

        drawdown = returns.expanding().max() - returns
        max_drawdown = drawdown.expanding().max()

        # Use (1-confidence_level) because the worst drawdowns are the biggest positive values
        max_drawdown_at_confidence_level = max_drawdown.quantile(1 - confidence_level, interpolation='higher')
        conditional_drawdown = np.nanmean(max_drawdown[max_drawdown >= max_drawdown_at_confidence_level])
        return conditional_drawdown



"""
Test functions from codependence module: correlation distances, mutual info, variation of information.
"""

import unittest
import numpy as np

from famlafl.codependence.correlation import (squared_angular_distance, angular_distance, absolute_angular_distance,
                                               distance_correlation)
from famlafl.codependence.information import (get_mutual_info, variation_of_information_score,
                                               get_optimal_number_of_bins)
from famlafl.codependence.codependence_matrix import (get_dependence_matrix, get_distance_matrix)
from famlafl.codependence.gnpr_distance import (spearmans_rho, gpr_distance, gnpr_distance)
from famlafl.util.generate_dataset import get_classification_data

# pylint: disable=invalid-name

class TestCodependence(unittest.TestCase):
    """
    Test codependence module.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        state = np.random.RandomState(42)
        self.x = state.normal(size=1000)
        self.y_1 = self.x ** 2 + state.normal(size=1000) / 5
        self.y_2 = abs(self.x) + state.normal(size=1000) / 5
        self.X_matrix, _ = get_classification_data(6, 2, 2, 100, sigma=0)

    def test_correlations(self):
        """
        Test correlation based coefficients: angular (abs, square), distance correlation.
        """
        angular_dist = angular_distance(self.x, self.y_1)
        sq_angular_dist = squared_angular_distance(self.x, self.y_1)
        abs_angular_dist = absolute_angular_distance(self.x, self.y_1)
        dist_corr = distance_correlation(self.x, self.y_1)

        self.assertAlmostEqual(angular_dist, 0.67, delta=1e-2)
        self.assertAlmostEqual(abs_angular_dist, 0.6703, delta=1e-2)
        self.assertAlmostEqual(sq_angular_dist, 0.7, delta=1e-2)
        self.assertAlmostEqual(dist_corr, 0.529, delta=1e-2)

        dist_corr_y_2 = distance_correlation(self.x, self.y_2)
        self.assertAlmostEqual(dist_corr_y_2, 0.5216, delta=1e-2)

    def test_information_metrics(self):
        """
        Test mutual info, information variability metrics.
        """
        mut_info = get_mutual_info(self.x, self.y_1, normalize=False)
        mut_info_norm = get_mutual_info(self.x, self.y_1, normalize=True)
        mut_info_bins = get_mutual_info(self.x, self.y_1, n_bins=10)

        # Test mutual info score
        self.assertAlmostEqual(mut_info, 0.522, delta=1e-2)
        self.assertAlmostEqual(mut_info_norm, 0.64, delta=1e-2)
        self.assertAlmostEqual(mut_info_bins, 0.626, delta=1e-2)

        # Test information variation score
        info_var = variation_of_information_score(self.x, self.y_1, normalize=False)
        info_var_norm = variation_of_information_score(self.x, self.y_1, normalize=True)
        info_var_bins = variation_of_information_score(self.x, self.y_1, n_bins=10)

        self.assertAlmostEqual(info_var, 1.4256, delta=1e-2)
        self.assertAlmostEqual(info_var_norm, 0.7316, delta=1e-2)
        self.assertAlmostEqual(info_var_bins, 1.418, delta=1e-2)

    def test_number_of_bins(self):
        """
        Test get_optimal_number_of_bins functions.
        """

        n_bins_x = get_optimal_number_of_bins(self.x.shape[0])
        n_bins_x_y = get_optimal_number_of_bins(self.x.shape[0], np.corrcoef(self.x, self.y_1)[0, 1])

        self.assertEqual(n_bins_x, 15)
        self.assertEqual(n_bins_x_y, 9)

    def test_codependence_matrix(self):
        '''
        Test the get_dependence_matrix and get_distance_matrix function
        '''

        # TODO: add tests for values in matrix
        #Dependence_matrix

        vi_matrix = get_dependence_matrix(self.X_matrix, dependence_method='information_variation')
        mi_matrix = get_dependence_matrix(self.X_matrix, dependence_method='mutual_information')
        corr_matrix = get_dependence_matrix(self.X_matrix, dependence_method='distance_correlation')
        rho_matrix = get_dependence_matrix(self.X_matrix, dependence_method='spearmans_rho')
        gpr_matrix = get_dependence_matrix(self.X_matrix, dependence_method='gpr_distance', theta=0.5)
        gnpr_matrix = get_dependence_matrix(self.X_matrix, dependence_method='gnpr_distance', theta=0.5, bandwidth=0.02)

        #Distance_matrix
        angl = get_distance_matrix(vi_matrix, distance_metric='angular')
        sq_angl = get_distance_matrix(mi_matrix, distance_metric='squared_angular')
        abs_angl = get_distance_matrix(corr_matrix, distance_metric='abs_angular')

        #assertions
        self.assertEqual(vi_matrix.shape[0], self.X_matrix.shape[1])
        self.assertEqual(mi_matrix.shape[0], self.X_matrix.shape[1])
        self.assertEqual(corr_matrix.shape[0], self.X_matrix.shape[1])
        self.assertEqual(rho_matrix.shape[0], self.X_matrix.shape[1])
        self.assertEqual(gpr_matrix.shape[0], self.X_matrix.shape[1])
        self.assertEqual(gnpr_matrix.shape[0], self.X_matrix.shape[1])

        self.assertEqual(angl.shape[0], self.X_matrix.shape[1])
        self.assertEqual(sq_angl.shape[0], self.X_matrix.shape[1])
        self.assertEqual(abs_angl.shape[0], self.X_matrix.shape[1])

    def test_value_error_raise(self):
        '''
        Test of invailid arguments
        '''
        #Unkown dependence_metric
        with self.assertRaises(ValueError):
            get_dependence_matrix(self.X_matrix, dependence_method='unknown')
        #Unkown distance_metric
        with self.assertRaises(ValueError):
            get_distance_matrix(self.X_matrix, distance_metric='unknown')

    def test_spearmans_rho(self):
        """
        Test spearmans_rho function.
        """

        rho_xy1 = spearmans_rho(self.x, self.y_1)
        rho_xy2 = spearmans_rho(self.x, self.y_2)

        self.assertAlmostEqual(rho_xy1, 0.0105586, delta=1e-7)
        self.assertAlmostEqual(rho_xy2, 0.0289523, delta=1e-7)

    def test_gpr_distance(self):
        """
        Test gnp_distance function.
        """

        gpr0_xy1 = gpr_distance(self.x, self.y_1, theta=0)
        gpr0_xy2 = gpr_distance(self.x, self.y_2, theta=0)

        gpr1_xy1 = gpr_distance(self.x, self.y_1, theta=1)
        gpr1_xy2 = gpr_distance(self.x, self.y_2, theta=1)

        self.assertAlmostEqual(gpr0_xy1, 0.3216183, delta=1e-7)
        self.assertAlmostEqual(gpr0_xy2, 0.3803020, delta=1e-7)

        self.assertAlmostEqual(gpr1_xy1, 0.7033639, delta=1e-7)
        self.assertAlmostEqual(gpr1_xy2, 0.6967954, delta=1e-7)

    def test_gnpr_distance(self):
        """
        Test gnpr_distance function.
        """

        gnpr0_xy1 = gnpr_distance(self.x, self.y_1, theta=0)
        gnpr0_xy2 = gnpr_distance(self.x, self.y_2, theta=0)

        gnpr1_xy1 = gnpr_distance(self.x, self.y_1, theta=1)
        gnpr1_xy2 = gnpr_distance(self.x, self.y_2, theta=1)

        self.assertAlmostEqual(gnpr0_xy1, 0.58834643, delta=1e-7)
        self.assertAlmostEqual(gnpr0_xy2, 0.57115983, delta=1e-7)

        self.assertAlmostEqual(gnpr1_xy1, 0.0032625, delta=1e-7)
        self.assertAlmostEqual(gnpr1_xy2, 0.0023459, delta=1e-7)



"""
Tests Robust Median Reversion.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.rmr import RMR


class TestRobustMedianReversion(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Robust Median Reversion class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_rmr_solution(self):
        """
        Test the calculation of RMR with the original method.
        """
        # Initialize RMR.
        rmr = RMR(epsilon=1.1, n_iteration=10, window=3, tau=0.001)
        # Allocates asset prices to RMR.
        rmr.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(rmr.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_rmr_epsilon_error(self):
        """
        Tests ValueError if epsilon is greater than 1.
        """
        # Initialize RMR.
        rmr1 = RMR(epsilon=0.5, n_iteration=10, window=3, tau=0.001)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr1.allocate(self.data)

    def test_rmr_n_iteration_error(self):
        """
        Tests ValueError if n_iteration is not an integer or less than 2.
        """
        # Initialize RMR.
        rmr2 = RMR(epsilon=1.2, n_iteration=1.5, window=3, tau=0.001)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr2.allocate(self.data)

        # Initialize RMR.
        rmr3 = RMR(epsilon=1.2, n_iteration=1, window=3, tau=0.001)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr3.allocate(self.data)

    def test_rmr_window_error(self):
        """
        Tests ValueError if window is not an integer or less than 2.
        """
        # Initialize RMR.
        rmr4 = RMR(epsilon=1.2, n_iteration=4, window=3.5, tau=0.001)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr4.allocate(self.data)

        # Initialize RMR.
        rmr5 = RMR(epsilon=1.2, n_iteration=4, window=1, tau=0.001)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr5.allocate(self.data)

    def test_rmr_break_solution(self):
        """
        Test the calculation of RMR with the break case in _calc_median.
        """
        # Initialize RMR.
        rmr6 = RMR(epsilon=1.1, n_iteration=10, window=3, tau=0.9)
        # Allocates asset prices to RMR.
        rmr6.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(rmr6.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_rmr_transform_non_mu(self):
        """
        Tests edge case for _transform non_mu edge case.
        """
        # pylint: disable=protected-access
        # pylint: disable=no-self-use
        # Initialize RMR.
        rmr7 = RMR(epsilon=1.1, n_iteration=10, window=3, tau=0.9)
        # Make an empty array.
        empty = np.zeros((1, 2))
        # Calculate edge case.
        rmr7._transform(empty, empty)

    def test_rmr_norm2_0_mu(self):
        """
        Tests edge case for norm2 = 0 in _transform method.
        """
        # Initialize RMR.
        rmr8 = RMR(epsilon=2, n_iteration=2, window=2, tau=0.9)
        # Make the data all ones.
        new_data = self.data
        new_data[:] = 1
        # Calculate edge case.
        rmr8.allocate(new_data, resample_by='M')

    def test_rmr_tau_error(self):
        """
        Tests ValueError if tau is less than 0 or greater than or equal to 1.
        """
        # Initialize RMR.
        rmr9 = RMR(epsilon=2, n_iteration=2, window=2, tau=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr9.allocate(self.data)

        # Initialize RMR.
        rmr10 = RMR(epsilon=2, n_iteration=2, window=2, tau=-1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            rmr10.allocate(self.data)



"""
Initialize Momentum Module.
"""
from famlafl.online_portfolio_selection.eg import EG
from famlafl.online_portfolio_selection.ftl import FTL
from famlafl.online_portfolio_selection.ftrl import FTRL



"""
Test various functions regarding chapter 5: Fractional Differentiation.
"""

import unittest
import os
import math
import numpy as np
import pandas as pd

from famlafl.features import fracdiff, plot_min_ffd


class TestFractionalDifferentiation(unittest.TestCase):
    """
    Test get_weights, get_weights_ffd, frac_diff, and frac_diff_ffd
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time')
        self.data.index = pd.to_datetime(self.data.index)

    def test_get_weights(self):
        """
        get_weights as implemented here matches the code in the book (Snippet 5.1).
        We test:
        1. if the first element of the weights vector is 1.0
        2. The length of the weights vector is same as the requested length
        """

        diff_amt = 0.9
        number_ele = 100
        weights = fracdiff.get_weights(diff_amt, size=number_ele)

        # Last value in the set is still the same
        self.assertTrue(weights[-1] == 1.0)

        # Size matches
        self.assertTrue(weights.shape[0] == number_ele)  # pylint: disable=unsubscriptable-object

    def test_get_weights_ffd(self):
        """
        get_weights_ffd as implemented here matches the code in the book (Snippet 5.2).
        We test:
        1. if the first element of the weights vector is 1.0
        2. The length of the weights vector is equal to 12

        """

        diff_amt = 0.9
        number_ele = 100
        thresh = 1e-3
        weights = fracdiff.get_weights_ffd(diff_amt, thresh=thresh, lim=number_ele)

        # Last value in the set is still the same
        self.assertTrue(weights[-1] == 1.0)

        # Size matches
        self.assertTrue(weights.shape[0] == 12)  # pylint: disable=unsubscriptable-object

    def test_frac_diff(self):
        """
        Assert that for any positive real number d,
        1. Length of the output is the same as the length of the input
        2. First element is NaN
        """
        data_series = self.data['close'].to_frame()

        for diff_amt in np.arange(0.1, 1, 0.1):
            fd_series = fracdiff.frac_diff(data_series, diff_amt=diff_amt)
            self.assertTrue(fd_series.shape[0] == len(data_series))
            self.assertTrue(isinstance(fd_series['close'].iloc[0], np.float64) and math.isnan(fd_series['close'].iloc[0]))

    def test_frac_diff_ffd(self):
        """
        Assert that for any positive real number d,
        1. Length of the output is the same as the length of the input
        2. First element is NaN
        """
        data_series = self.data['close'].to_frame()

        for diff_amt in np.arange(0.1, 1, 0.1):
            fd_series = fracdiff.frac_diff_ffd(data_series, diff_amt=diff_amt)
            self.assertTrue(fd_series.shape[0] == len(data_series))
            self.assertTrue(isinstance(fd_series['close'].iloc[0], np.float64) and math.isnan(fd_series['close'].iloc[0]))

    def test_plot_min_ffd(self):
        """
        Assert that the plot for min ffd is correct,

        Testing is based on the correlation between the original series (d=0)
        and the differentiated series.
        """
        data_series = self.data['close'].to_frame()

        expected_correlation = np.array([[0, 1],
                                         [0.1, 0.99295323],
                                         [0.2, 0.97712122],
                                         [0.3, 0.95098824],
                                         [0.4, 0.91422650],
                                         [0.5, 0.86412240],
                                         [0.6, 0.80909555],
                                         [0.7, 0.76457507],
                                         [0.8, 0.67228154],
                                         [0.9, 0.62583193],
                                         [1.0, 0.51058195]])

        # Obtaining a plot for min ffd
        plot = plot_min_ffd(data_series)
        correlation = plot.lines[0].get_xydata()

        print(type(plot))


        # Test if equal
        np.testing.assert_allclose(correlation, expected_correlation)



"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.data_structures import time_data_structures as ds


class TestTimeDataStructures(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Dollar bars
    2. Volume bars
    3. Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/tick_data_time_bars.csv'

    def test_day_bars(self):
        """
        Tests the seconds bars implementation.
        """

        db1 = ds.get_time_bars(self.path, resolution='D', num_units=1, batch_size=1000, verbose=False)
        db2 = ds.get_time_bars(self.path, resolution='D', num_units=1, batch_size=50, verbose=False)
        db3 = ds.get_time_bars(self.path, resolution='D', num_units=1, batch_size=10, verbose=False)
        ds.get_time_bars(self.path, resolution='D', num_units=1, batch_size=50, verbose=False,
                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv')

        # Assert diff batch sizes have same number of bars
        self.assertEqual(db1.shape[0], 1)
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db1.shape == db4.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db1.values == db4.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[0, 'open'] == 1200.0)
        self.assertTrue(db1.loc[0, 'high'] == 1249.75)
        self.assertTrue(db1.loc[0, 'low'] == 1200.0)
        self.assertTrue(db1.loc[0, 'close'] == 1249.75)

        # Assert date_time is correct
        self.assertTrue(db1.loc[0, 'date_time'] == pd.Timestamp(2011, 8, 1, 0, 0, 0).timestamp())

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_hour_bars(self):
        """
        Tests the seconds bars implementation.
        """

        db1 = ds.get_time_bars(self.path, resolution='H', num_units=1, batch_size=1000, verbose=False)
        db2 = ds.get_time_bars(self.path, resolution='H', num_units=1, batch_size=50, verbose=False)
        db3 = ds.get_time_bars(self.path, resolution='H', num_units=1, batch_size=10, verbose=False)
        ds.get_time_bars(self.path, resolution='H', num_units=1, batch_size=50, verbose=False,
                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv')

        # Assert diff batch sizes have same number of bars
        self.assertEqual(db1.shape[0], 3)
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[1, 'open'] == 1225.0)
        self.assertTrue(db1.loc[1, 'high'] == 1249.75)
        self.assertTrue(db1.loc[1, 'low'] == 1225.0)
        self.assertTrue(db1.loc[1, 'close'] == 1249.75)

        # Assert date_time is correct
        self.assertTrue(db1.loc[1, 'date_time'] == pd.Timestamp(2011, 8, 1, 0, 0, 0).timestamp())

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_minute_bars(self):
        """
        Tests the minute bars implementation.
        """

        db1 = ds.get_time_bars(self.path, resolution='MIN', num_units=1, batch_size=1000, verbose=False)
        db2 = ds.get_time_bars(self.path, resolution='MIN', num_units=1, batch_size=50, verbose=False)
        db3 = ds.get_time_bars(self.path, resolution='MIN', num_units=1, batch_size=10, verbose=False)
        ds.get_time_bars(self.path, resolution='MIN', num_units=1, batch_size=50, verbose=False,
                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv')

        # Assert diff batch sizes have same number of bars
        self.assertEqual(db1.shape[0], 11)
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[9, 'open'] == 1275.0)
        self.assertTrue(db1.loc[9, 'high'] == 1277.0)
        self.assertTrue(db1.loc[9, 'low'] == 1275.0)
        self.assertTrue(db1.loc[9, 'close'] == 1277.0)

        # Assert date_time is correct
        self.assertTrue(db1.loc[9, 'date_time'] == pd.Timestamp(2011, 8, 1, 23, 39, 0).timestamp())

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_second_bars(self):
        """
        Tests the seconds bars implementation.
        """

        db1 = ds.get_time_bars(self.path, resolution='S', num_units=10, batch_size=1000, verbose=False)
        db2 = ds.get_time_bars(self.path, resolution='S', num_units=10, batch_size=50, verbose=False)
        db3 = ds.get_time_bars(self.path, resolution='S', num_units=10, batch_size=10, verbose=False)
        ds.get_time_bars(self.path, resolution='S', num_units=10, batch_size=50, verbose=False,
                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv')

        # Assert diff batch sizes have same number of bars
        self.assertEqual(db1.shape[0], 47)
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[1, 'open'] == 1201.0)
        self.assertTrue(db1.loc[1, 'high'] == 1202.0)
        self.assertTrue(db1.loc[1, 'low'] == 1201.0)
        self.assertTrue(db1.loc[1, 'close'] == 1202.0)

        # Assert date_time is correct
        self.assertTrue(db1.loc[1, 'date_time'] == pd.Timestamp(2011, 7, 31, 22, 39, 0).timestamp())

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_wrong_input_value_error_raise(self):
        """
        Tests ValueError raise when neither pd.DataFrame nor path to csv file are passed to function call
        """
        with self.assertRaises(ValueError):
            ds.get_time_bars(None, resolution='MIN', num_units=1, batch_size=1000, verbose=False)

    def test_csv_format(self):
        """
        Asserts that the csv data being passed is of the correct format.
        """
        wrong_date = ['2019-41-30', 200.00, np.int64(5)]
        wrong_price = ['2019-01-30', 'asd', np.int64(5)]
        wrong_volume = ['2019-01-30', 200.00, '1.5']
        too_many_cols = ['2019-01-30', 200.00, np.int64(5), 'Limit order', 'B23']

        # pylint: disable=protected-access
        self.assertRaises(ValueError, ds.TimeBars._assert_csv,
                          pd.DataFrame(wrong_date).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.TimeBars._assert_csv,
                          pd.DataFrame(too_many_cols).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.TimeBars._assert_csv,
                          pd.DataFrame(wrong_price).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.TimeBars._assert_csv,
                          pd.DataFrame(wrong_volume).T)



"""
Entropy calculation module (Shannon, Lempel-Ziv, Plug-In, Konto)
"""

import math
from typing import Union

import numpy as np
from numba import njit


def get_shannon_entropy(message: str) -> float:
    """
    Advances in Financial Machine Learning, page 263-264.

    Get Shannon entropy from message

    :param message: (str) Encoded message
    :return: (float) Shannon entropy
    """
    exr = {}
    entropy = 0
    for each in message:
        try:
            exr[each] += 1
        except KeyError:
            exr[each] = 1
    textlen = len(message)
    for value in exr.values():
        freq = 1.0 * value / textlen
        entropy += freq * math.log(freq) / math.log(2)
    entropy *= -1
    return entropy


def get_lempel_ziv_entropy(message: str) -> float:
    """
    Advances in Financial Machine Learning, Snippet 18.2, page 266.

    Get Lempel-Ziv entropy estimate

    :param message: (str) Encoded message
    :return: (float) Lempel-Ziv entropy
    """
    i, lib = 1, [message[0]]
    while i < len(message):
        for j in range(i, len(message)):
            message_ = message[i:j + 1]
            if message_ not in lib:
                lib.append(message_)
                break
        i = j + 1
    return len(lib) / len(message)


def _prob_mass_function(message: str, word_length: int) -> dict:
    """
    Advances in Financial Machine Learning, Snippet 18.1, page 266.

    Compute probability mass function for a one-dim discete rv

    :param message: (str or array) Encoded message
    :param word_length: (int) Approximate word length
    :return: (dict) Dict of pmf for each word from message
    """
    lib = {}
    if not isinstance(message, str):
        message = ''.join(map(str, message))
    for i in range(word_length, len(message)):
        message_ = message[i - word_length:i]
        if message_ not in lib:
            lib[message_] = [i - word_length]
        else:
            lib[message_] = lib[message_] + [i - word_length]
    pmf = float(len(message) - word_length)
    pmf = {i: len(lib[i]) / pmf for i in lib}
    return pmf


def get_plug_in_entropy(message: str, word_length: int = None) -> float:
    """
    Advances in Financial Machine Learning, Snippet 18.1, page 265.

    Get Plug-in entropy estimator

    :param message: (str or array) Encoded message
    :param word_length: (int) Approximate word length
    :return: (float) Plug-in entropy
    """
    if word_length is None:
        word_length = 1
    pmf = _prob_mass_function(message, word_length)
    out = -sum([pmf[i] * np.log2(pmf[i]) for i in pmf]) / word_length
    return out


@njit()
def _match_length(message: str, start_index: int, window: int) -> Union[int, str]:    # pragma: no cover
    """
    Advances in Financial Machine Learning, Snippet 18.3, page 267.

    Function That Computes the Length of the Longest Match

    :param message: (str or array) Encoded message
    :param start_index: (int) Start index for search
    :param window: (int) Window length
    :return: (int, str) Match length and matched string
    """
    # Maximum matched length+1, with overlap.
    sub_str = ''
    for length in range(window):
        msg1 = message[start_index: start_index + length + 1]
        for j in range(start_index - window, start_index):
            msg0 = message[j: j + length + 1]
            if len(msg1) != len(msg0):
                continue
            if msg1 == msg0:
                sub_str = msg1
                break  # Search for higher l.
    return len(sub_str) + 1, sub_str  # Matched length + 1


def get_konto_entropy(message: str, window: int = 0) -> float:
    """
    Advances in Financial Machine Learning, Snippet 18.4, page 268.

    Implementations of Algorithms Discussed in Gao et al.[2008]

    Get Kontoyiannis entropy

    :param message: (str or array) Encoded message
    :param window: (int) Expanding window length, can be negative
    :return: (float) Kontoyiannis entropy
    """
    out = {
        'h': 0,
        'r': 0,
        'num': 0,
        'sum': 0,
        'sub_str': []
    }
    if window <= 0:
        points = range(1, len(message) // 2 + 1)
    else:
        window = min(window, len(message) // 2)
        points = range(window, len(message) - window + 1)
    for i in points:
        if window <= 0:
            length, msg_ = _match_length(message, i, i)
            out['sum'] += np.log2(i + 1) / length  # To avoid Doeblin condition
        else:
            length, msg_ = _match_length(message, i, window)
            out['sum'] += np.log2(window + 1) / length  # To avoid Doeblin condition
        out['sub_str'].append(msg_)
        out['num'] += 1
    try:
        out['h'] = out['sum'] / out['num']
    except ZeroDivisionError:
        out['h'] = 0
    out['r'] = 1 - out['h'] / (np.log2(len(message)) if np.log2(len(message)) > 0 else 1)  # Redundancy, 0<=r<=1
    return out['h']



# pylint: disable=missing-module-docstring
import numpy as np
import scipy.optimize as opt
from famlafl.online_portfolio_selection.corn import CORN


class SCORN(CORN):
    """
    This class implements the Symmetric Correlation Driven Nonparametric Learning strategy. It is
    reproduced with modification from the following paper:
    `Wang, Y., & Wang, D. (2019). Market Symmetry and Its Application to Pattern-Matching-Based
    Portfolio Selection. The Journal of Financial Data Science, 1(2), 78–92.
    <https://jfds.pm-research.com/content/1/2/78>`_

    Symmetric Correlation Driven Nonparametric Learning is an extension of the CORN strategy
    proposed by Yang Wang and Dong Wang. SCORN looks to not only maximize the returns for the similar
    periods but also minimize the losses from the negatively correlated periods.
    """

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Create similar set.
        similar_set = []

        # Creat opposite set.
        opposite_set = []

        # Default is uniform weights.
        new_weights = self._uniform_weight()

        # Calculate for similar sets if time is greater or equal to window size.
        if time >= self.window:
            # Iterate through past windows.
            for past_time in range(time - self.window + 1):
                # Check for windows that are above rho.

                if self.corr_coef[time - self.window + 1][past_time] > self.rho:
                    # Append the time for similar set.
                    similar_set.append(past_time + self.window)

                # Check for windows that are below rho.
                elif self.corr_coef[time - self.window + 1][past_time] < -self.rho:
                    # Append the time for opposite set.
                    opposite_set.append(past_time + self.window)

            # If either set exists, continue to return new weights.
            if similar_set or opposite_set:
                # Choose the corresponding relative return periods.
                similar = self.relative_return[similar_set]
                opposite = self.relative_return[opposite_set]
                new_weights = self._scorn_optimize(similar, opposite)
        return new_weights

    def _scorn_optimize(self, similar, opposite):
        """
        Calculates weights that maximize returns over the given array.

        :param similar: (np.array) Relative returns of similar periods.
        :param opposite: (np.array) Relative returns of inversely similar periods.
        :return: (np.array) Weights that maximize the returns for the given array.
        """
        # Initialize guess.
        weights = self._uniform_weight()

        # Use np.log and np.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        def _objective(weight):
            return -np.sum(np.log(np.dot(similar, weight))) + np.sum(np.log(np.dot(opposite, weight)))

        # Derivative of the objective function.
        def _derivative(weight):
            similar_returns = np.dot(similar, weight)
            opposite_returns = np.dot(opposite, weight)
            return -np.dot(1 / similar_returns, similar) + np.dot(1 / opposite_returns, opposite)

        # Weight bounds.
        bounds = tuple((0.0, 1.0) for asset in range(self.number_of_assets))

        # Sum of weights is 1.
        const = ({'type': 'eq', 'fun': lambda w: np.sum(w) - 1})

        problem = opt.minimize(_objective, weights, method='SLSQP', bounds=bounds, constraints=const, jac=_derivative)
        return problem.x



"""
Tests the unit functions in ef3m.py for calculating fitting a mixture of 2 Gaussian distributions.
"""

import unittest
import numpy as np
import pandas as pd
from scipy.special import comb

from famlafl.bet_sizing.ef3m import (M2N, centered_moment, raw_moment, most_likely_parameters, iter_4_jit, iter_5_jit)


class TestM2NConstructor(unittest.TestCase):
    """
    Tests the constructor method of the M2N class.
    """

    def test_m2n_constructor(self):
        """
        Tests that the constructor of the M2N class executes properly.
        """
        moments_test = [1, 2, 3, 4, 5]
        m2n_test = M2N(moments_test)
        # Confirm that the initial attributes get set properly.
        self.assertEqual(m2n_test.moments, moments_test)
        self.assertEqual(m2n_test.new_moments, [0, 0, 0, 0, 0])
        self.assertEqual(m2n_test.parameters, [0, 0, 0, 0, 0])
        self.assertEqual(m2n_test.error, sum([moments_test[i]**2 for i in range(len(moments_test))]))


class TestM2NGetMoments(unittest.TestCase):
    """
    Tests the 'get_moments' method of the M2N class.
    """
    def test_get_moments(self):
        """
        Tests the 'get_moments' method of the M2N class.
        """
        u_1, u_2, s_1, s_2, p_1 = [2.1, 4.3, 1.1, 0.7, 0.3]
        p_2 = 1 - p_1
        m_1 = p_1*u_1 + p_2*u_2
        m_2 = p_1*(s_1**2 + u_1**2) + p_2*(s_2**2 + u_2**2)
        m_3 = p_1*(3*s_1**2*u_1 + u_1**3) + p_2*(3*s_2**2*u_2 + u_2**3)
        m_4 = p_1*(3*s_1**4 + 6*s_1**2*u_1**2 + u_1**4) + p_2*(3*s_2**4 + 6*s_2**2*u_2**2 + u_2**4)
        m_5 = p_1*(15*s_1**4*u_1 + 10*s_1**2*u_1**3 + u_1**5) + p_2*(15*s_2**4*u_2 + 10*s_2**2*u_2**3 + u_2**5)
        test_params = [u_1, u_2, s_1, s_2, p_1]
        test_mmnts = [m_1, m_2, m_3, m_4, m_5]
        # Create M2N object.
        m2n_test = M2N(test_mmnts)
        # Check self-return method.
        m2n_test.get_moments(test_params, return_result=False)
        self.assertEqual(test_mmnts, m2n_test.new_moments)
        # Check the function when 'return_value' is True.
        result_mmnts = m2n_test.get_moments(test_params, return_result=True)
        self.assertEqual(test_mmnts, result_mmnts)


class TestM2NIter4(unittest.TestCase):
    """
    Tests the 'iter_4' method of the M2N class.
    """
    def test_iter_4_validity_check_1(self):
        """
        Tests 'iter_4' method's 'Validity check 1' breakpoint condition.
        """
        moments_test = [1, 2, 3, 4, 5]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(3, 1)
        self.assertTrue(not param_results)

    def test_iter_4_validity_check_2(self):
        """
        Tests 'iter_4' method's 'Validity check 2' breakpoint condition.
        """
        moments_test = [2, 2, 3, 4, 5]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(1, 0.8)
        self.assertTrue(not param_results)

    def test_iter_4_validity_check_3(self):
        """
        Tests 'iter_4' method's 'Validity check 3' breakpoint condition.
        """
        moments_test = [1.5, 2, 3, 4, 5]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(2, 0.7)
        self.assertTrue(not param_results)

    # Validity check 4 is not present since the condition appears to be unreachable.

    def test_iter_4_validity_check_5(self):
        """
        Tests 'iter_4' method's 'Validity check 5' breakpoint condition.
        """
        moments_test = [0.0, 0.1, 0.0, 0.0, 5]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(0.1, 0.5)
        self.assertTrue(not param_results)

    def test_iter_4_validity_check_6(self):
        """
        Tests 'iter_4' method's 'Validity check 6' breakpoint condition.
        """
        moments_test = [0.0, 0.1, 0.0, 0.0, 5]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(0.1, 0.25)
        self.assertTrue(not param_results)

    def test_iter_4_success(self):
        """
        Tests 'iter_4' method for successful execution.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_4(1, 0.2)
        self.assertTrue(len(param_results) == 5)


class TestM2NIter5(unittest.TestCase):
    """
    Tests the 'iter_5' method of the M2N class.
    """
    def test_iter_5_validity_check_1(self):
        """
        Tests 'iter_5' method's 'Validity check 1' breakpoint condition.
        """
        moments_test = [0.0, 0.0, 0.0, 0.0, 0.0]
        mu_2_test, p_1_test = 0.0, 0.05
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_2(self):
        """
        Tests 'iter_5' method's 'Validity check 2' breakpoint condition.
        """
        moments_test = [0.0, 0.0, 0.0, 0.0, 0.0]
        mu_2_test, p_1_test = 0.1, 0.05
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_3(self):
        """
        Tests 'iter_5' method's 'Validity check 3' breakpoint condition.
        """
        moments_test = [0.0, 0.0, 0.1, 0.0, 0.0]
        mu_2_test, p_1_test = 0.1, 0.2
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    # Validity check 4 is not present since the condition appears to be unreachable.

    def test_iter_5_validity_check_5(self):
        """
        Tests 'iter_5' method's 'Validity check 5' breakpoint condition.
        """
        moments_test = [0.0, 0.1, 0.0, 0.0, 0.0]
        mu_2_test, p_1_test = 0.1, 0.99999
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_6(self):
        """
        Tests 'iter_5' method's 'Validity check 6' breakpoint condition.
        """
        moments_test = [0.0, 0.1, 0.0, 0.0, 0.0]
        mu_2_test, p_1_test = 0.1, 0.95
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_7(self):
        """
        Tests 'iter_5' method's 'Validity check 7' breakpoint condition.
        """
        moments_test = [0.0, 0.1, 0.1, 0.0, 0.2]
        mu_2_test, p_1_test = 0.4, 0.95
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_8(self):
        """
        Tests 'iter_5' method's 'Validity check 8' breakpoint condition.
        """
        moments_test = [1.7486117351052706, 12.30094642908807, 44.14804719610457, 301.66990880582324, 1389.7073066865096]
        mu_2_test, p_1_test = 8.927498436080297, -1910484717784700.2
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_validity_check_9(self):
        """
        Tests 'iter_5' method's 'Validity check 9' breakpoint condition.
        """
        moments_test = [1.7465392043495434, 12.32010406019726, 44.3090981635415, 302.3152423573811, 1403.0640473698527]
        mu_2_test, p_1_test = 1.8733475857864539, 0.019291066689915537
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(not param_results)

    def test_iter_5_success(self):
        """
        Tests 'iter_5' method for successful execution.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test, p_1_test = 0.8642146104188053, 0.03296760034110158
        m2n_test = M2N(moments_test)
        param_results = m2n_test.iter_5(mu_2_test, p_1_test)
        self.assertTrue(len(param_results) == 5)


class TestM2NFit(unittest.TestCase):
    """
    Tets the 'fit' method of the M2N class.
    """
    def test_fit_variant_1(self):
        """
        Tests the 'fit' method of the M2N class, using variant 1.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test = 1
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 5
        variant_test = 1
        max_iter_test = 10_000
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        m2n_test.fit(mu_2_test)
        self.assertTrue(len(m2n_test.parameters) == 5)

    def test_fit_variant_2(self):
        """
        Tests the 'fit' method of the M2N class, using variant 2.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test = 1
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 5
        variant_test = 2
        max_iter_test = 10_000
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        m2n_test.fit(mu_2_test)
        self.assertTrue(len(m2n_test.parameters) == 5)

    def test_fit_variant_value_error(self):
        """
        Tests that the 'fit' method throws a ValueError if an invalid value is passed to argument 'variant'.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test = 1
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 5
        variant_test = 3
        max_iter_test = 10_000
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        self.assertRaises(ValueError, m2n_test.fit, mu_2_test)

    def test_fit_success_via_error(self):
        """
        Tests that the 'fit' method successfully exits due to a low error being reached.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test = 1
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 5
        variant_test = 1
        max_iter_test = 10_000
        mu_2_test, epsilon_test, variant_test, max_iter_test = 1, 1e-5, 1, 10_000
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        m2n_test.error = 1e6
        m2n_test.fit(mu_2_test)
        self.assertTrue(len(m2n_test.parameters) == 5)

    def test_fit_success_via_epsilon(self):
        """
        Tests that the 'fit' method successfully exits due to p_1 converging.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        mu_2_test = 1
        epsilon_test = 1e12
        factor_test = 5
        n_runs_test = 5
        variant_test = 1
        max_iter_test = 10_000
        mu_2_test, epsilon_test, variant_test, max_iter_test = 1, 1e12, 1, 10_000
        np.random.seed(12)
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        m2n_test.fit(mu_2_test)
        self.assertTrue(len(m2n_test.parameters) == 5)

    def test_fit_success_via_max_iter(self):
        """
        Tests that the 'fit' method successfully exits due to the maximum number of iterations being reached.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        np.random.seed(12)
        mu_2_test = 1
        epsilon_test = 1e-12
        factor_test = 5
        n_runs_test = 5
        variant_test = 1
        max_iter_test = 1
        mu_2_test, epsilon_test, variant_test, max_iter_test = 1, 1e-12, 1, 1
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        m2n_test.fit(mu_2=mu_2_test)
        self.assertTrue(len(m2n_test.parameters) == 5)


class TestM2NEF3M(unittest.TestCase):
    """
    Tests the EF3M algorithms of the M2N module.
    """

    def test_ef3m_variant_1(self):
        """
        Tests the 'iter_4_jit' function of the M2N module (using variant 1).
        """
        mu_2 = -0.5876905479546004
        p_1 = 0.020950069267730465
        moments_test = [
            -0.59,
            9.830000000000002,
            -19.922000000000004,
            264.254,
        ]
        expected_parameters = [
            -0.6979265579594388,
            -0.5876905479546004,
            20.80638279572574,
            0.6488995559493259,
            0.0004662589937450317,
        ]
        param_list = iter_4_jit(mu_2, p_1, *moments_test)
        param_list = param_list.tolist()
        self.assertTrue(expected_parameters == param_list)

    def test_ef3m_variant_2(self):
        """
        Tests the 'iter_5_jit' function of the M2N module (using variant 2).
        """
        mu_2 = -0.07037328978510915
        p_1 = 0.09166890087206325
        moments_test = [
            -0.59,
            9.830000000000002,
            -19.922000000000004,
            264.254,
            -818.2100000000003,
        ]
        expected_parameters = [
            -5.738890150700704,
            0.1755905244444409,
            0.862038546663441,
            2.723656932895561,
            0.12317856407155195,
        ]
        param_list = iter_5_jit(mu_2, p_1, *moments_test)
        param_list = param_list.tolist()
        self.assertTrue(expected_parameters == param_list)


class TestM2NSingleFitLoop(unittest.TestCase):
    """
    Tests the 'single_fit_loop' method.
    """
    def test_single_fit_loop_return_type(self):
        """
        Tests that the 'single_fit_loop' method executes successfully.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 10
        variant_test = 2
        max_iter_test = 10_000
        epsilon_test, factor_test, variant_test, max_iter_test = 1e-5, 5, 2, 10_000
        np.random.seed(12)
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test)
        df_results = m2n_test.single_fit_loop()
        self.assertTrue(isinstance(df_results, pd.DataFrame))


class TestM2NMpFit(unittest.TestCase):
    """
    Tests the 'mp_fit' method.
    """
    def test_mp_fit_return_type(self):
        """
        Tests that the 'mp_fit' method executes successfully.
        """
        moments_test = [0.7, 2.6, 0.4, 25, -59.8]
        epsilon_test = 1e-5
        factor_test = 5
        n_runs_test = 10
        variant_test = 2
        max_iter_test = 10_000
        num_workers_test = 1
        epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test, num_workers_test = 1e-5, 5, 10, 2, 10_000, 1
        m2n_test = M2N(moments_test, epsilon_test, factor_test, n_runs_test, variant_test, max_iter_test, num_workers_test)
        df_results = m2n_test.mp_fit()
        self.assertTrue(isinstance(df_results, pd.DataFrame))


class TestCenteredMoment(unittest.TestCase):
    """
    Tests the helper function 'centered_moment'.
    """
    def test_centered_moment_result(self):
        """
        Tests for the successful execution of the 'centered_moment' helper function.
        """
        raw_test = [0.701756, 2.591815, 0.450519, 24.689030, -57.756735]
        centered_5th_correct = 0
        for j in range(6):
            if j == 5:
                add_on = 1
            else:
                add_on = raw_test[5-j-1]
            centered_5th_correct += (-1)**j * int(comb(5, j)) * add_on * raw_test[0]**j
        centered_5th_test = centered_moment(raw_test, 5)
        self.assertAlmostEqual(centered_5th_test, centered_5th_correct, 7)


class TestRawMoment(unittest.TestCase):
    """
    Tests the helper function 'raw_moment'.
    """
    def test_raw_moment_result(self):
        """
        Tests for the successful execution of the 'raw_moment' helper function.
        """
        centered_test = [0.0, 2.11, -4.373999999999999, 30.803699999999996, -153.58572]
        raw_result = raw_moment(centered_test, 0.7)
        raw_correct = [0.7, 2.6, 0.4, 25, -59.8]
        self.assertTrue(np.allclose(raw_result, raw_correct, 1e-7))


class TestMostLikelyParameters(unittest.TestCase):
    """
    Tests the helper function 'most_likely_parameters'.
    """
    def test_most_likely_parameters_result(self):
        """
        Tests for the successful execution of the 'most_likely_parameters' function.
        """
        mu_1_list = [-2.074149682208028, -2.1464760973734522, -1.7318027625411423, -1.7799163398785354, -1.9766582333677596]
        mu_2_list = [0.9958122958772418, 0.9927128514876395, 1.013574632526087, 1.0065707257309104, 1.009533655971151]
        sigma_1_list = [1.9764097851543956, 1.9516780127056625, 2.080573657129795, 2.071328499049906, 1.9988591140726848]
        sigma_2_list = [1.002964090440232, 1.0054392587806025, 0.9872577865302316, 0.9909001363163131, 0.9971327048101786]
        p_1_list = [0.09668610445835334, 0.09379917992315062, 0.11351960785118335, 0.10993400151299484, 0.10264463363929438]
        df_test = pd.DataFrame.from_dict({'mu_1': mu_1_list,
                                          'mu_2': mu_2_list,
                                          'sigma_1': sigma_1_list,
                                          'sigma_2': sigma_2_list,
                                          'p_1': p_1_list})
        most_likely_correct = {'mu_1': -2.03765,
                               'mu_2': 1.00863,
                               'sigma_1': 1.9832,
                               'sigma_2': 1.00012,
                               'p_1': 0.09942}
        d_results = most_likely_parameters(data=df_test)
        self.assertTrue(np.allclose(list(d_results.values()), list(most_likely_correct.values()), 1e-7))

    def test_most_likely_parameters_list_arg(self):
        """
        Tests the helper function 'most_likely_parameters' when passing a list to 'ignore_columns'.
        """
        mu_1_list = [-2.074149682208028, -2.1464760973734522, -1.7318027625411423, -1.7799163398785354, -1.9766582333677596]
        mu_2_list = [0.9958122958772418, 0.9927128514876395, 1.013574632526087, 1.0065707257309104, 1.009533655971151]
        sigma_1_list = [1.9764097851543956, 1.9516780127056625, 2.080573657129795, 2.071328499049906, 1.9988591140726848]
        sigma_2_list = [1.002964090440232, 1.0054392587806025, 0.9872577865302316, 0.9909001363163131, 0.9971327048101786]
        p_1_list = [0.09668610445835334, 0.09379917992315062, 0.11351960785118335, 0.10993400151299484, 0.10264463363929438]
        df_test = pd.DataFrame.from_dict({'mu_1': mu_1_list,
                                          'mu_2': mu_2_list,
                                          'sigma_1': sigma_1_list,
                                          'sigma_2': sigma_2_list,
                                          'p_1': p_1_list})
        most_likely_correct = {'mu_1': -2.03765,
                               'mu_2': 1.00863,
                               'sigma_1': 1.9832,
                               'sigma_2': 1.00012,
                               'p_1': 0.09942}
        d_results = most_likely_parameters(data=df_test, ignore_columns=['error'])
        self.assertTrue(np.allclose(list(d_results.values()), list(most_likely_correct.values()), 1e-7))



# pylint: disable=missing-module-docstring
import numpy as np
import scipy.optimize as opt
from famlafl.online_portfolio_selection.ftl import FTL


class FTRL(FTL):
    """
    This class implements the Follow the Regularized Leader strategy. It is reproduced with
    modification from the following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    Follow the Regularized Leader strategy directly tracks the Best Constant Rebalanced Portfolio
    until the previous period with an additional regularization term
    """

    def __init__(self, beta):
        """
        Initializes Follow the Regularized Leader with a beta constant term.

        :param beta: (float) Constant to the regularization term. Typical ranges for interesting
                             results include [0, 0.2], 1, and any high values. Low beta
                             FTRL strategies are identical to FTL, and high beta indicates more
                             regularization to return a uniform CRP.
        """
        super(FTRL, self).__init__()
        self.beta = beta

    def _fast_optimize(self, optimize_array):
        """
        Calculates weights that maximize returns over the given array.

        :param optimize_array: (np.array) Relative returns of the assets for a given time period.
        :return: (np.array) Weights that maximize the returns for the given array.
        """
        # Initialize guess.
        weights = self._uniform_weight()

        # Use np.log and np.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        # Add additional l2 regularization term for the weights for calculation.
        def _objective(weight):
            return -np.sum(np.log(np.dot(optimize_array, weight))) + self.beta * np.linalg.norm(weight) / 2

        # Weight bounds.
        bounds = tuple((0.0, 1.0) for asset in range(self.number_of_assets))

        # Sum of weights is 1.
        const = ({'type': 'eq', 'fun': lambda w: np.sum(w) - 1})

        problem = opt.minimize(_objective, weights, method='SLSQP', bounds=bounds, constraints=const)
        return problem.x
